// SL 23-07-2011: New compiler based on the new scanner. See comments in Scanner.cs
using System.Collections.Generic;
using System.Collections;
using System.Reflection;
using System;
using System.Text;
using System.IO;
using System.Xml.XPath;
using VisTool;
using System.Xml;
using System.Diagnostics;

namespace VisTool {

    public partial class Parser {
        // Error handling: If the form file exists, OpenVisFile will always return a form template. If the file doesn't exist, null is returned.
        // Errors when creating a child template will create a dummy template (Label).
        // Errors when creating a property will skip the property. Example: Property doesn't start with indentifier + colon.

        public static FormTemplate OpenVisFile(string frmName1) { // The relative file name without .vis
            // Loads the vis-file into a form template, but doesn't compile it.
            Parser parser;
            StreamReader str;
            Directory.SetCurrentDirectory(VisEngine.ApplicationDirectory);
            if (!Parser.IsFormName(ref frmName1)) 
                return null; // Also adds the name to Forms and returns the name with true caps.

            //if (!System.IO.File.Exists(frmName1 + ".vis")) {
                //ErrorLogging.Log(ErrorMode.CompileTime, Priority.MsgBox, null, null, null, "Tried to open " + frmName1 + ".vis", null, null);
                //return null;
            //}
            using (str = new StreamReader(frmName1 + ".vis")) {
                parser = new Parser(str, false);

                frmName = tplName = frmName1;
                parser.tpl = parser.frm = new FormTemplate(frmName);
                //VisEngine.s2 = VisEngine.sw1.ElapsedMilliseconds;
                parser.tplProp = null; // For error messages: No property yet
                parser.ReadVisFile(); // Read and scan the vis-file (make template, properties and tokens). Don't compile yet.
                // Debug.Print("End of parsing vis-file. Create template tree\n");
                str.Close(); // Close StreamReader
                return parser.frm;
            }
        }

        // VisFile = [comment EOL] TemplateSpec { ComponentType: tplName [Ignore] EOL TemplateSpec } EOF
        // Read the file and generate templates, properties and tokens.
        // Space stuff before an initial EOL is comments about the template. Is appended to the form name in the name property.
        // Errors in the ComponentType line can not be reported until the template and the name proporty have been created.
        // In case of errors, a dummy name and/or a label component are created. The original line appears as a comment.
        void ReadVisFile() {
            string comment = null; // Define the name property for the Form.
            tplProp = tpl.valPropArray[VisComponent.NameID];
            tplProp.Formula = frmName.TrimEnd(); 
            if (la.kind == TkKind.EOL) {
                if (!String.IsNullOrWhiteSpace(la.space)) {
                    tplProp.startChain = null; // Scan again.
                    frm.Comment = la.space.TrimEnd(); 
                    tplProp.Formula = frmName + " " + frm.Comment; // Make comment visible as comment to Form name. 
                }
                Get(); // Read EOL
            }
            try {
                TemplateSpec(frm); // Form template. Ends with separator line or EOF. Token = EOL

                while (!la.EOF) { // ComponentType: tplName EOL TemplateSpec
                    string ctrlType = null, tplName = "Dummy", nameError = null; // Default names to prepare for initial errors. 
                    Token firstToken = la;
                    StartOut(la); // Prepare for errors by making the entire line a comment. 
                    comment = token.space.Trim(); // Space before first token is template comments.
                    if (la.kind != TkKind.Ident) {
                        nameError = "Expects a component type"; goto create;  // Default will be used
                    }
                    ctrlType = la.val; // Component type
                    Get(); if (la.val != ":") { nameError = "Expects colon"; goto create; }
                    Get(); if (la.kind != TkKind.Ident) { nameError = "Expects name"; goto create; } // Default will be used.
                    tplName = la.val;
                    if (frm.Templates.ContainsKey(tplName)) { nameError = "Template name \"" + tplName + "\" used already"; goto create; }
                    PropInfo propInfo;
                    if (TplProperty.GetTemplateProperty(frm, tplName, out propInfo) != null) nameError = "Template name \"" + tplName + "\" is also a Form property"; 
create:
                    if (nameError != null) { // There was an error. Find a unique template name
                        int i = 0; string s = tplName;
                        while (frm.Templates.ContainsKey(s) || TplProperty.GetTemplateProperty(frm, s, out propInfo) != null ) 
                            s = tplName + "_" + (++i).ToString();
                        tplName = s;
                    }
                    tpl = null;
                    if (ctrlType.ToUpper() == "VISFORM")
                        nameError = "A VisForm cannot be on a Form";
                    else {
                        tpl = CreateTemplate((ctrlType == null ? "Label" : ctrlType), tplName, frm); // Create template.
                        if (tpl == null) nameError = "Component type \"" + ctrlType + "\" not found";
                    }
                    if (tpl == null || ctrlType == null) {
                        tpl = CreateTemplate("Label", tplName, frm); // Create dummy template
                        VisEngine.CreateOrSetTemplatePropertyNoLock(tpl, "BackColor", "Color.Yellow");
                        VisEngine.CreateOrSetTemplatePropertyNoLock(tpl, "Text", "\"" + firstToken.val + " ?\"");
                    }
                    tpl.Comment = comment;

                    // This is a non-form template. Name is also a property with comments
                    string formula;
                    if (nameError != null) { // Create a two-token Name formula: dummy("", tplName) EOL(' Was:, the original component type line)
                        while (la.kind != TkKind.EOL) Get(); // The rest of the line becomes a comment.
                        Get(); // Get EOL
                        // Insert dummy token at the head of the chain followed by the EOL.
                        firstToken = new Token(lineCount); firstToken.next = token;
                        firstToken.kind = TkKind.Ident; firstToken.space = ""; firstToken.val = tplName;
                        token.space = " \' Was: "; token.val = bOut.ToString().Trim();
                        formula = tplName + " \' Was: " + bOut.ToString().Trim();
                    } else {
                        StartOut(La); // Make the part from name to EOL a formula 
                        firstToken = la; Get();  // Get name.
                        if (la.val.ToUpper() == "IGNORE") {
                            tpl.Ignore = true; Get();
                        }
                        if (la.kind != TkKind.EOL)
                            error("Unexpected symbol", la);
                        while (la.kind != TkKind.EOL) Get(); // The rest of the line is copied to the formula.
                        Get(); // Get EOL
                        formula = bOut.ToString(); //.Trim();
                    }
                    tplProp = tpl.valPropArray[VisComponent.NameID];
                    tplProp.Formula = formula.Trim();
                    //tplProp.InstancePropIndex = tpl.DesignProperties.Count;

                    tplProp.startChain = firstToken;
                    token.next = token; // Terminate the formula with an endless loop of EOL

                    // We now have a template and a property suitable for error logging
                    if (nameError != null) {
                        error(nameError, firstToken.next);
                        //ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, tpl, tplProp, nameError, firstToken.next.val, firstToken.next);
                    }
                    frm.Templates[tpl.Name] = tpl;
                    tpl.FormTpl = frm;
                    TemplateSpec(tpl); // Has read EOL as end
                } // End while
                frm.FormTpl.isDirty = false;
            }
            catch (Exception e) {
                frm.FormTpl.isDirty = false;
                error(e.Message, token);
                //ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, tpl, tplProp, e.Message, token.val, token);
            }
        } // End VisFile

        // TemplateSpec = { propertyName : [ parentID |  {formulaToken} ] EOL } [ - - {-} EOL | EOF ]   Note: Empty formula allowed.
        // Errors in the property name or colon can not be reported until the property has been created.
        // In case of errors, an Error Property is created. The original line appears as a comment in its formula.
        void TemplateSpec(Template tplOrFrm) {
            PropInfo propInfo = null;
            while (!(la.val == "--" || la.EOF)) { // Until separator line or EOF
                string propName = "Dummy", nameError = null; // Default name to prepare for initial errors.
                Token firstToken = la; // The first token of the ComponentType line.
                StartOut(la); // Prepare for errors by making the entire line a comment. 
                //try {
                if (la.kind != TkKind.Ident) { nameError = "Expects property name"; goto create; } // Default will be used
                propName = la.val;
                Get(); if (la.val != ":") { nameError = "Expects colon"; goto create; }
                Get(); // Get colon

                TplProperty tplProp = TplProperty.GetTemplateProperty(tplOrFrm, propName, out propInfo);
                if (propName.ToUpper() == "NAME") {
                    nameError = "The template name must be defined in this way: Label: lblName";
                } else if (propName.ToUpper() == "INDEX") {
                    nameError = "Index is read-only"; 
                } else if (propName.ToUpper() == "LAST") {
                    nameError = "Last is read-only";
                } else if (tplOrFrm == frm && propName.ToUpper() == "FORM") {
                    nameError = "The Form name is not a property. It is the form's file name";
                } else if (tplProp != null && (tplProp.Formula != null || propInfo == null)) { // Property exists with a formula or as a designer property.
                    string tail = ""; if (propName == "BackColor" || propName == "Text") tail = " (maybe to indicate an error)";
                    nameError = "Property name \"" + propName + "\" used already" + tail;
                } else if (propInfo != null && propInfo.VisAttribute.ReadOnly && la.kind != TkKind.EOL) { // Comment allowed for read-only.
                        nameError = "Property name \"" + propName + "\" is read-only";
                }
create: // Create an error property or the real property.
                string formula;
                if (nameError != null) { // Create this space and val in EOL:  ' Was: OriginalPropertyName: formula
                    // Find a unique property name:
                    int i = 1; string s = "ErrorProperty_" + "1";
                    while (TplProperty.GetTemplateProperty(tplOrFrm, s, out propInfo) != null) s = "ErrorProperty_" + (++i).ToString();
                    propName = s;
                    tplProp = VisEngine.CreateOrSetTemplatePropertyNoLock(tplOrFrm, propName, null); // The formula will be set below
                    while (la.kind != TkKind.EOL) Get(); // The rest of the line becomes a comment.
                    Get(); // Get EOL
                    tplProp.startChain = firstToken = token;
                    // Insert space and val in the EOL token.
                    token.space = " \' Was: "; token.val = bOut.ToString().Trim();
                    token.lines = 0; // The modified token.space has no \n.
                    formula = " \' Was: " + bOut.ToString().Trim();
                    // The property can now receive error logging.
                    error(nameError, token);
                    //ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, tpl, tplProp, nameError, token.val, token);
                } else {
                    tplProp = VisEngine.CreateOrSetTemplatePropertyNoLock(tplOrFrm, propName, null); // The formula will be set below
                    tplProp.startChain = firstToken = la;
                    // The property can now receive error logging.
                    StartOut(la); // Make only the part after colon into a formula.
                    if (la.space.StartsWith(" ")) la.space = la.space.Remove(0, 1);
                    // Remove one space if present. Leave other initial whitespace.
                    while (la.kind != TkKind.EOL) {
                        Get(); if (token.error != null)
                            error(token.error, token); // Report error detected by scanner
                    }
                    Get(); // Get EOL
                    formula = bOut.ToString(); // Don't trim. EOL has no trailing whitespace except for a space for error marking.
                }
                tplProp.Formula = formula.TrimEnd();
                token.next = token; // Terminate the formula with an endless loop of EOL.
                token.EOF = true; // This is End-of-File for the formula.
            } // end while. All properties read.

            // Assert: la.val == "--" || la.EOF
            Token errorToken = token;
            while (la.val == "--" || la.val == "-") Get(); // Read separator line
            if (la.kind != TkKind.EOL) {
                error("A separator line must consist of only --- (minuses)", errorToken);
                while (la.kind != TkKind.EOL) Get(); // Skip the rest of the separator line.
            }
            if (la.error != null) error(la.error, la);
            Get(); // Get EOL
        } // end TemplateSpec

        /// <summary>
        /// Create a template. Returns null if the template name isn't recognized..
        /// </summary>
        /// <param name="tplType">The component type that the template should create, e.g.: TextBox</param>
        /// <param name="tplName">The name for the template</param>
        /// <param name="frmTpl">The form template for the template being created</param>
        /// <returns></returns>
        public static Template CreateTemplate(string tplType, string tplName, FormTemplate frmTpl) {
            ComponentInfo info;
            if (VisEngine.ComponentByName.TryGetValue(tplType, out info))
                return new Template(tplName, info, frmTpl);
            return null; // Do nothing?? The caller will handle it as "Bad component type".
        }
    } // End class Compiler
} // end namespace VisTool.Kernel
