﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Windows.Forms;
using System.Diagnostics; //Trace = System.Diagnostics.Trace;
using vbSystemFunctions;
using VisTool; 
// using vbSystemFunctions;


namespace VisTool {

    public partial class SystemItems {
        public static PopupForm Popup; // Holds most recent popup form.
        public static VisComponent PopupSource; // The component that opened the popup. ClosePopup will only 
        // close the popup if it is the same source as the one that opened it. This is needed because the MouseEnter 
        // event is followed by a MouseLeave from the previous component. 

        // System Variables
        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "X position of the mouse relative to my canvas",
            ShortHelp = "X position of the mouse relative to my canvas", Position = 20)]
        public static int MouseX=5;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "Y position of the mouse relative to my canvas",
            ShortHelp = "Y position of the mouse relative to my canvas", Position = 20)]
        public static int MouseY = 5;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "The mouse button pressed",
            ShortHelp = "The mouse button pressed", Position = 21)]
        public static MouseButtons MouseButton = System.Windows.Forms.MouseButtons.None;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "Number of ticks with the mouse wheel",
            ShortHelp = "Number of ticks with the mouse wheel", Position = 22)]
        public static int MouseDelta = 0;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "My Left before move started.",
            ShortHelp = "My Left before move started.", Position = 23)]
        public static int OldLeft = 0;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "My Width before resize started.",
            ShortHelp = "My Width before resize started.", Position = 24)]
        public static int OldWidth=0;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "My Top before move started.",
            ShortHelp = "My Top before move started.", Position = 25)]
        public static int OldTop=0;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "My Height before resize started.",
            ShortHelp = "My Height before resize started.", Position = 26)]
        public static int OldHeight=0;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, ShortHelp = "The new value to be stored at data entry.", Position = 27)]
        public static object NewValue = null;

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod,
            LongHelp = "(c) True if component c has keyboard focus (receives all keyboard events).",
            ShortHelp = "(c) True if component c has keyboard focus (receives all keyboard events).", Position = 29)]
        public static bool GotFocus(VisComponent component) {
            return component == VisEngine.SelectedComponent;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "The key pressed. See list of keys with System.Keys",
            ShortHelp = "The key pressed. See list of keys with System.Keys", Position = 30)]
        public static Keys Key = System.Windows.Forms.Keys.None;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "True when Ctrl is down",
            ShortHelp = "True when Ctrl is down", Position = 31)]
        public static bool CtrlPressed = false;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "True when Alt is down",
            ShortHelp = "True when Alt is down", Position = 31)]
        public static bool AltPressed = false;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "True when Shift is down",
            ShortHelp = "True when Shift is down", Position = 31)]
        public static bool ShiftPressed = false;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, 
            LongHelp = @"When true, don't look at other event handlers for this event. Initially false.
In FormClosing, set Handled to true to cancel closing.",
    ShortHelp = "When true, don't look at other event handlers for this event. Initially false.", Position = 32)]
        public static bool Handled = false;

        public static FieldInfo CheckVariable(string varName) { // Used by PathEngine to access system properties.
            FieldInfo fieldInfo = typeof(SystemItems).GetField(varName, System.Reflection.BindingFlags.IgnoreCase | System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.Static);
            return fieldInfo;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Field, ShortHelp = "The height of the primary screen excluding taskbar, etc.", Position = 33)]
        public static int ScreenHeight = System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Height;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, ShortHelp = "The width of the primary screen.", Position = 34)]
        public static int ScreenWidth = System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Width;

        [VisSystemAtt(VisSystemAtt.ItemType.Field, ShortHelp = "The Zoom factor, i.e. 1.3 or 0.8. Default 1.0.", Position = 27)]
        public static float Zoom = 1.0F; // Accessed directly by System.Zoom.
        public static float Zoom2 { // Used internally to check for and correct negative zoom factors.
            get { if (Zoom < 0) Zoom = 1.0F; return Zoom; }
            //set { if (value > 0) Zoom = value; }
        }
        /* 
 * Formulas compute sizes and positions in logical pixels. UVis multiplies sizes and positions with Zoom2 to define physical screen pixels. 
 * Zoom can be defined in the vism file or set by the designer with System.Zoom = x. If we make Zoom a get/set field, reflection cannot 
 * access it. So we cannot check that x is positive. Access through Zoom2 makes the check. 
 * Text points are in principle physical sizes. One point is 1/72 inch, i.e. 0.353 mm. For screens with high DPI (Dots Per Inch), 
 *  a text point should occupy more pixels to get the correct physical size. You can get the actual DPI of the primary screen from Window's Graphics objects. 
  * However, all screens seem to have 96 DPI, whether they physically have large or small pixels.
  * In principle, texts and graphics should scale differently to preserve the physical size of the letters, but this may wreck the layout. 
  * So for simplicity, we scale everything with the Zoom factor. For documentation,uVis prints some scaling data in the errorlog file. 
 */

        //Enum Types

        [VisSystemAtt(VisSystemAtt.ItemType.UvisEnum, LongHelp = "List of predefined colors. Construct others with ColorRGB and ColorARGB",
            ShortHelp = "List of predefined colors. Construct others with ColorRGB and ColorARGB")]
        public static object Color(string colorName, out List<string> Suggestions) {  //Get the suggestions if colorName is null.
            object Color = null;
            Suggestions = null;
            Type colorType = typeof(System.Drawing.Color);
            List<string> suggestions = new List<string>();
            if (colorName != null) {
                if (colorType.GetProperty(colorName, BindingFlags.IgnoreCase | BindingFlags.Static | BindingFlags.DeclaredOnly | BindingFlags.Public) == null)
                    return null; // Seems the only way to test that color is known. IgnoreCase alone is useless
                Color = System.Drawing.Color.FromName(colorName);
            } else {
                // We take only static property to avoid properties like Name, IsSystemColor ...
                PropertyInfo[] propInfos = colorType.GetProperties(BindingFlags.Static | BindingFlags.DeclaredOnly | BindingFlags.Public);
                foreach (PropertyInfo propInfo in propInfos)
                    suggestions.Add(propInfo.Name);
                Suggestions = suggestions;
            }
            return Color;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.UvisEnum, LongHelp = "List of text fonts",
            ShortHelp = "List of text fonts")]
        public static object Font(string fontName, out List<string> Suggestions) {  //Get the suggestions if colorName is null.
            Suggestions = null;
            Type colorType = typeof(System.Drawing.Color);
            List<string> suggestions = new List<string>();
            if (fontName != null) {
                return fontName; // Font names are handled as strings.
            } else {
                System.Drawing.FontFamily[] fontArray = System.Drawing.FontFamily.Families;
                Suggestions = new List<string>(fontArray.Length);                
                foreach (var font in fontArray)
                    Suggestions.Add(font.Name);
            }
            return null;
        }

        //[VisEnum]
        //public static object Math(string propName, out List<string> suggestions) {
        //    object math = null;
        //    suggestions = new List<string>();
        //    if (propName != null) {
        //        FieldInfo fieldInfo = typeof(System.Math).GetField(propName, System.Reflection.BindingFlags.IgnoreCase | System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.Static);
        //        if (fieldInfo != null)
        //            math = fieldInfo.GetValue(null);
        //    } else {
        //        Type mathType = typeof(System.Math);
        //        // We take only static property to avoid properties like Name, IsSystemColor ...
        //        FieldInfo[] fieldInfos = mathType.GetFields(BindingFlags.Static | BindingFlags.DeclaredOnly | BindingFlags.Public);
        //        foreach (FieldInfo fieldInfo in fieldInfos)
        //            suggestions.Add(fieldInfo.Name);
        //    }
        //    return math;
        //}

        [VisSystemAtt(VisSystemAtt.ItemType.UvisEnum, LongHelp = "List of keyboard Keys that may be down",
            ShortHelp = "List of keyboard Keys that may be down")]
        public static object Keys(string propName, out List<string> suggestions) {
            object keys = null;
            suggestions = new List<string>();
            if (propName != null) {
                FieldInfo fieldInfo = typeof(Keys).GetField(propName, System.Reflection.BindingFlags.IgnoreCase | System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.Static);
                if (fieldInfo != null)
                    keys = fieldInfo.GetValue(null);
            } else {
                // Use only static properties to avoid properties like Name, IsSystemColor ...
                FieldInfo[] fieldInfos = typeof(Keys).GetFields(BindingFlags.Static | BindingFlags.DeclaredOnly | BindingFlags.Public);
                foreach (FieldInfo fieldInfo in fieldInfos)
                    suggestions.Add(fieldInfo.Name);
            }
            return keys;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.UvisEnum, LongHelp = "List of mouse buttons that may be down",
            ShortHelp = "List of mouse buttons that may be down")]
        public static object MouseButtons(string propName, out List<string> suggestions) {
            object mouseButtons = null;
            suggestions = new List<string>();
            if (propName != null) {
                FieldInfo fieldInfo = typeof(MouseButtons).GetField(propName, System.Reflection.BindingFlags.IgnoreCase | System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.Static);
                if (fieldInfo != null)
                    mouseButtons = fieldInfo.GetValue(null);
            } else {
                // We take only static property to avoid properties like Name, IsSystemColor ...
                FieldInfo[] fieldInfos = typeof(MouseButtons).GetFields(BindingFlags.Static | BindingFlags.DeclaredOnly | BindingFlags.Public);
                foreach (FieldInfo fieldInfo in fieldInfos)
                    suggestions.Add(fieldInfo.Name);
            }
            return mouseButtons;
        }
        // End enum

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, 
            LongHelp = "(a, r, g, b) Color with transparency (a), red (r), green (g), blue (b). Values 0-255.",
            ShortHelp = "(a, r, g, b) Color with transparency (a), red (r), green (g), blue (b).", Position = 40)]
        public static System.Drawing.Color? ColorARGB(object alpha, object red, object green, object blue) {
            if (alpha==null || red == null || green == null || blue == null) return null;
            return System.Drawing.Color.FromArgb(Convert.ToInt32(alpha), Convert.ToInt32(red), Convert.ToInt32(green), Convert.ToInt32(blue)); 
            // Errors will be handled by Eval and Default.
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, 
            LongHelp = "(r, g, b) Color with red (r), green (g), blue (b). Values 0-255.",
            ShortHelp = "(r, g, b) Color with red (r), green (g), blue (b).", Position = 40)]
        public static System.Drawing.Color? ColorRGB(object red, object green, object blue) {
            if (red == null || green == null || blue == null) return null;
            return System.Drawing.Color.FromArgb(255, Convert.ToInt32(red), Convert.ToInt32(green), Convert.ToInt32(blue)); 
            // Errors will be handled by Eval and Default.
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod,
            ShortHelp = "(n) The n'th color in a sequence of colors as distinct as possible.", Position = 40)]
        public static System.Drawing.Color? ColorByNumber(int number) { // Experimental
            int n = number ^ (number >> 1); // The Gray code for number. Changes only one bit when number increases by one.
            int red = 0, green = 0, blue = 0, colorBit = 128;
            for (int i = 0; i < 8; i++) {
                if (n % 2 == 0) red += colorBit;
                n = n >> 1;
                if (n % 2 == 0) green += colorBit;
                n = n >> 1;
                if (n % 2 == 0) blue += colorBit;
                n = n >> 1;
                colorBit = colorBit >> 1;
            };
            Debug.Print("Color " + number + ": " + red + ", " + green + ", " + blue);
            return System.Drawing.Color.FromArgb(255, red > 255 ? 255 : red, green > 255 ? 255 : green, blue > 255 ? 255 : blue);
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(n) Convert n (number or string) to byte (0-255).",
            ShortHelp = "(n) Convert n (number or string) to byte (0-255).", Position = 41)]
        public static Byte? CByte(object s) { if (s == null) return null; return Convert.ToByte(s); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(n) Convert n (number or string) to short integer (16 bit).\nThe Access database calls this \"integer\".",
            ShortHelp = "(n) Convert n (number or string) to short integer (16 bit).", Position = 42)]
        public static int? CShort(object s) { if (s == null) return null; return Convert.ToInt16(s); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(n) Convert n (number or string) to integer (32 bit).\nThe Access database calls this \"long\".",
            ShortHelp = "(n) Convert n (number or string) to integer (32 bit).", Position = 43)]
        public static int? CInt(object s) { if (s == null) return null; return Convert.ToInt32(s); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod,
            LongHelp = "(n) Convert n (number or string) to integer (64 bit).\nThe Access database doesn't support this integer.",
            ShortHelp = "(n) Convert n (number or string) to integer (64 bit).", Position = 44)]
        public static Int64? CLng(object s) { if (s == null) return null; return Convert.ToInt64(s); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(n) Round n (number or string) to closest double-precision, floating-point integer.",
            ShortHelp = "(n) Round n (number or string) to closest double-precision, floating-point integer.", Position = 44)]
        public static double? Round(object s) { if (s == null) return null; return System.Math.Round(Convert.ToDouble(s)); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(n) Convert n (number or string) to a double-precision, floating-point number.",
            ShortHelp = "(n) Convert n (number or string) to a double-precision, floating-point number.", Position = 45)]
        public static double? CDbl(object s) { if (s == null) return null; return Convert.ToDouble(s); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s) Convert string s to DateTime.",
            ShortHelp = "(s) Convert string s to DateTime.", Position = 46)]
        public static DateTime? CDate(object s) { if (s == null) return null; return Convert.ToDateTime(s); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(a) True if a is null.",
            ShortHelp = "(a) True if a is null.", Position = 47)]
        public static bool IsNull(object e) { return e == null; }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = false,
            LongHelp = "(a) True if a is null or a string of white space.",
            ShortHelp = "(a) True if a is null or a string of white space.", Position = 47)]
        public static bool IsBlank(object e) {
            if (e == null) return true;
            return (e is string && String.IsNullOrWhiteSpace((string)e));
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s) True if s can be converted to 32-bit integer.",
            ShortHelp = "(s) True if s can be converted to 32-bit integer.", Position = 47)]
        public static bool IsInt(object s) {
            try { Convert.ToInt32(s); } catch (Exception) { return false; }
            return true;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s) True if s can be converted to 64-bit integer.",
            ShortHelp = "(s) True if s can be converted to 64-bit integer.", Position = 47)]
        public static bool IsLng(object s) {
            try { Convert.ToInt64(s); } catch (Exception) { return false; }
            return true;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s) True if s can be converted to a double-precision, floating point number.",
            ShortHelp = "(s) True if s can be converted to a double-precision, floating point number.", Position = 47)]
        public static bool IsDbl(object s) {
            try { Convert.ToDouble(s); } catch (Exception) { return false; }
            return true;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s) True if s can be converted to DateTime.",
            ShortHelp = "(s) True if s can be converted to DateTime.", Position = 47)]
        public static bool IsDate(object s) {
            try { Convert.ToDateTime(s); } catch (Exception) { return false; }
            return true;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(dtn) Convert dtn (DateTime or number) to string. Use Format(dtn, f) for advanced formatting.",
            ShortHelp = "(dtn) Convert dtn (DateTime or number) to string.", Position = 48)]
        public static string CStr(object o) { if (o == null) return null; return o.ToString(); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(dtn, f) Convert dtn (DateTime or number) to a string, using a format string, e.g. f=\"0.00\" or \"yyyy-MM-dd\"",
            ShortHelp = "(dtn, f) Convert dtn (DateTime or number) to a string, using a format string, e.g. f=\"0.00\" or \"yyyy-MM-dd\"", Position = 48)]
        public static string Format(object obj, string format) {
            if (obj == null || format == null) return null;
            //return ((double)obj).ToString(format); Must cast each type to allow format.
            return vbSystemFunctions.vbFunctions.vbFormat(obj, format);
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true, Position = 50,
            ShortHelp = @"(n) The n'th character in the alphabet as a string, e.g. Chr(65)=""A"".
Chr(13)+Chr(10) is new line, but Chr(10) often works alone. Use NewLine() instead.")]
        public static string Chr(object i) { if (i == null) return null; return Convert.ToChar(i).ToString(); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s) Number of characters in string s.",
            ShortHelp = "(s) Number of characters in string s.", Position = 51)]
        public static int? Len(object o) { if (o == null) return null; return o.ToString().Length; }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s, n) The first n characters of string s.",
            ShortHelp = "(s, n) The first n characters of string s.", Position = 52)]
        public static string Left(object o, object count) {
            if (o == null || count == null) return null;
            string s = o.ToString();
            int c = Convert.ToInt16(count);
            if (s.Length <= c) return s;
            return s.Substring(0, c);
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s, n) The last n characters of string s.",
            ShortHelp = "(s, n) The last n characters of string s.", Position = 53)]
        public static string Right(object o, object count) {
            if (o == null || count == null) return null;
            string s = o.ToString();
            int c = Convert.ToInt16(count);
            if (s.Length <= c) return s;
            return s.Substring(s.Length - c, c);
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s, m, n) n characters of string s, starting with character m.\nMid(\"abcd\", 2, 1) = \"b\" ",
            ShortHelp = "(s, m, n) n characters of string s, starting with character m.", Position = 54)]
        public static string Mid(object o, object from, object count) { // VB string has 1-based char indexes
            if (o == null || from == null || count == null) return null;
            string s = o.ToString();
            int f = Convert.ToInt16(from);
            int c = Convert.ToInt16(count);
            if (s.Length < f + c) return s.Substring(f - 1);
            return s.Substring(f - 1, c);
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s) String s without leading spaces.",
            ShortHelp = "(s) String s without leading spaces.", Position = 55)]
        public static string LTrim(object o) { if (o == null) return null; return o.ToString().TrimStart(); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s) String s without trailing spaces.",
            ShortHelp = "(s) String s without trailing spaces.", Position = 56)]
        public static string RTrim(object o) { if (o == null) return null; return o.ToString().TrimEnd(); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, 
            LongHelp = "(s) String s without leading or trailing spaces.",
            ShortHelp = "(s) String s without leading or trailing spaces.", Position = 57)]
        public static string Trim(object o) { if (o == null) return null; return o.ToString().Trim(); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s) String s with all letters in lower case.",
            ShortHelp = "(s) String s with all letters in lower case.", Position = 58)]
        public static string Lcase(object o) { if (o == null) return null; return o.ToString().ToLowerInvariant(); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(s) String s with all letters in upper case.",
            ShortHelp = "(s) String s with all letters in upper case.", Position = 58)]
        public static string Ucase(object o) { if (o == null) return null; return o.ToString().ToUpperInvariant(); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(n) A string with n spaces.",
             ShortHelp = "(n) A string with n spaces.", Position = 60)]
        public static string Space(object count) {
            if (count == null) return null;
            int c = Convert.ToInt16(count);
            string s = "";
            for (int i = 0; i < c; i++) s += " "; // Use string builder ??
            return s;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true, Position = 60,
            ShortHelp = "( ) A string with a new-line character. Actually \\r\\n")]
        public static string NewLine() { return "\r\n"; }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(b, c, d) When b is true, get c, otherwise get d (immediate if). UVis Evaluates only c or d.",
            ShortHelp = "(b, c, d) When b is true, get c, otherwise get d (immediate if). UVis Evaluates only c or d.", Position = 10)]
        public static object iif(object b, Expression e1, Expression e2) { 
            if (b == null) return null;
            return Convert.ToBoolean(b) ? e1.Eval(VisEngine.curComponent, VisEngine.curTemplate) : 
                e2.Eval(VisEngine.curComponent, VisEngine.curTemplate);
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(n, v1, v2 ...) Get the n'th parameter. n from 1 and up.\nUVis Evaluates only the n'th parameter.",
            ShortHelp = "(n, v1, v2 ...) Get the n'th parameter. n from 1 and up.", Position = 10)]
        public static object Choose(object j, params Expression [] c) {
            if (j == null) return null;
            int i = Convert.ToInt32(j);
            if (i < 1 || i > c.Length) {
                Token t = VisEngine.curExprToken;
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                    "Index is " + i + ". Must be between 1 and " + c.Length, t, VisEngine.curComponent.Index());
                throw new LoggedException();
            } else
                return c[i - 1].Eval(VisEngine.curComponent, VisEngine.curTemplate);
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, // Although this is SQL-supported, uVis may simulate the time, i.e. evaluate it as a constant.
            ShortHelp = "( ) Current date and time. Simulated time when in simulation mode.", Position = 79)]
        public static DateTime Now() { return RealOrSimulatedDateTime(); } // Visual Basic 6.0

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, // Although this is SQL-supported, uVis may simulate the time, i.e. evaluate it as a constant.
            ShortHelp = "(dt) Sets simulated time to dt. Returns true when in simulation mode.", Position = 80)]
        public static bool SetNow(DateTime dt) { VisEngine.simulatedDateTime = dt; return VisEngine.simulatedTime; } // Visual Basic 6.0

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, // Although this is SQL-supported, uVis may simulate the time, i.e. evaluate it as a constant.
            LongHelp = "( ) Current date at 0:00. Simulated time when in simulation mode.",
            ShortHelp = "( ) Current date at 0:00. Simulated time when in simulation mode.", Position = 81)]
        public static DateTime Date() { return RealOrSimulatedDateTime().Date; } // Visual Basic 6.0

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod,
            LongHelp = "( ) Current date at 0:00. Simulated time when in simulation mode.Same as Date().\nNot supported by SQL",
            ShortHelp = "( ) Current date at 0:00. Simulated time when in simulation mode.", Position = 81)]
        public static DateTime Today() { return RealOrSimulatedDateTime().Date; } // Visual Basic .NET

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, 
            LongHelp = "( ) Current time since midnight. Simulated time when in simulation mode. Same as Time().\nNot supported by SQL",
            ShortHelp = "( ) Current time since midnight. Simulated time when in simulation mode. Same as Time().", Position = 82)]
        public static TimeSpan TimeOfDay() { return RealOrSimulatedDateTime().TimeOfDay; } // Visual Basic .NET

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, // Although this is SQL-supported, uVis may simulate the time, i.e. evaluate it as a constant.
            LongHelp = "( ) Current time since midnight. Simulated time when in simulation mode.",
            ShortHelp = "( ) Current time since midnight. Simulated time when in simulation mode.", Position = 82)]
        public static TimeSpan Time() { return RealOrSimulatedDateTime().TimeOfDay; } // Visual Basic 6.0

        static DateTime RealOrSimulatedDateTime() {
            if (VisEngine.simulatedTime)
                return VisEngine.simulatedDateTime;
            else
                return DateTime.Now;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, 
            LongHelp = "(f, incr, dt) Add incr years (f=\"y\"), incr months (f=\"m\"), days (f=\"d\"), \nhours (f=\"h\"), minutes (f=\"n\") or seconds (f=\"s\") to DateTime dt.",
            ShortHelp = "(f, incr, dt) Add incr years (f=\"y\"), incr months (f=\"m\"), etc. to DateTime dt.", Position = 99)]
        public static DateTime? DateAdd(string format, object increment, object baseDate) {
            if (format == null || increment == null || baseDate == null) return null;
            DateTime d = (DateTime)CDate(baseDate);
            double dbl = Convert.ToDouble(increment);
            switch (format) {
            case "y": //year
                d = d.AddYears((int)dbl);
                break;
            case "m": //year
                d = d.AddMonths((int)dbl);
                break;
            case "d": //day
                d = d.AddDays(dbl);
                break;
            case "h": //hours
                d = d.AddHours(dbl);
                break;
            case "n": //minute
                d = d.AddMinutes(dbl);
                break;
            case "s": //second
                d = d.AddSeconds(dbl);
                break;
            }
            return d;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true, 
            LongHelp = "(dt) The day number of DateTime dt.",
            ShortHelp = "(dt) The day number of DateTime dt.", Position = 92)]
        public static int? Day(object d) { if (d == null) return null; return Convert.ToDateTime(d).Day; }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(dt) The month number of DateTime dt.",
            ShortHelp = "(dt) The month number of DateTime dt.", Position = 91)]
        public static int? Month(object d) { if (d == null) return null; return Convert.ToDateTime(d).Month; }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true, 
            LongHelp = "(dt) The year of DateTime dt.",
            ShortHelp = "(dt) The year of DateTime dt.", Position = 90)]
        public static int? Year(object d) { if (d == null) return null; return Convert.ToDateTime(d).Year; }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(dt) The week-day number of DateTime dt. Sunday is 0.",
            ShortHelp = "(dt) The week-day number of DateTime dt. Sunday is 0.", Position = 93)]
        public static int? WeekDay(object d) { if (d == null) return null; return (int)Convert.ToDateTime(d).DayOfWeek; }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(dt) The hour of DateTime dt.",
            ShortHelp = "(dt) The hour of DateTime dt.", Position = 94)]
        public static int? Hour(object d) { if (d == null) return null; return Convert.ToDateTime(d).Hour; }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(dt) The minute of DateTime dt.",
            ShortHelp = "(dt) The minute of DateTime dt.", Position = 95)]
        public static int? Minute(object d) { if (d == null) return null; return Convert.ToDateTime(d).Minute; }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(dt) The second of DateTime dt.",
            ShortHelp = "(dt) The second of DateTime dt.", Position = 96)]
        public static int? Second(object d) { if (d == null) return null; return Convert.ToDateTime(d).Second; }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "( ) The name of the user who is logged in.",
            ShortHelp = "( ) The name of the user who is logged in.", Position = 98)]
        public static string CurrentUserName() { return System.Environment.UserName; }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "( ) The domain name of the user who is logged in.",
            ShortHelp = "( ) The domain name of the user who is logged in.", Position = 99)]
        public static string CurrentUserDomainName() { return System.Environment.UserDomainName; }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, // Not supported by Access. Should support a^b
            ShortHelp = "(x, y) The larger of x and y. May be number, dateTime or string.", Position = 103)]
        public static object MaxVal(object x, object y) { 
            if (x == null || y == null) return null;
            if (x is DateTime && y is DateTime) return (DateTime)x > (DateTime)y ? x : y;
            if (x is string && y is string) return (String.Compare((x as string), (y as string), true) > 0 ? x : y);
            if (x is int && y is int) return (int)x > (int)y ? x : y; 
            return System.Math.Max(Convert.ToDouble(x), Convert.ToDouble(y)); 
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, // Not supported by Access. Should support a^b
            ShortHelp = "(x, y) The smaller of x and y. May be number, dateTime or string.", Position = 103)]
        public static object MinVal(object x, object y) { 
            if (x == null || y == null) return null; 
            if (x is DateTime && y is DateTime) return (DateTime)x < (DateTime)y ? x : y;
            if (x is string && y is string) return (String.Compare((x as string), (y as string), true) < 0 ? x : y);
            if (x is int && y is int) return (int)x < (int)y ? x : y;
            double res = System.Math.Min(Convert.ToDouble(x), Convert.ToDouble(y));
            return res; 
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Field, LongHelp = "π = 3.141592 ... Circle circumference / diameter",
            ShortHelp = "π = 3.141592 ... Circle circumference / diameter", Position = 120)]
        public static double Pi = Math.PI;

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(x) The square root of x, e.g. Sqr(25) = 5.",
            ShortHelp = "(x) The square root of x.", Position = 121)]
        public static double? Sqr(object o) { if (o == null) return null; return System.Math.Sqrt(Convert.ToDouble(o)); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(n, b) The logarithm of n with base b, e.g. Log(100, 10) = 2.",
            ShortHelp = "(n, b) The logarithm of n with base b, e.g. Log(100, 10) = 2.", Position = 122)]
        public static double? Log(object a, object newBase) { if (a == null || newBase == null) return null; return System.Math.Log(Convert.ToDouble(a), Convert.ToDouble(newBase)); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, // Not supported by Access. Should support a^b
            LongHelp = "(x, y) x raised to the power of y, e.g. Pow(2, 10) = 1024.",
            ShortHelp = "(x, y) x raised to the power of y.", Position = 123)]
        public static double? Pow(object x, object y) { if (x == null || y == null) return null; return System.Math.Pow(Convert.ToDouble(x), Convert.ToDouble(y)); } 

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(r) The sine of angle r (in radian). r = degrees*Pi/180.",
            ShortHelp = "(r) The sine of angle r (in radian). r = degrees*Pi/180.", Position = 125)]
        public static double? Sin(object d) { if (d == null) return null; return System.Math.Sin(Convert.ToDouble(d)); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(x) The angle (in radian) whose sine is x.",
            ShortHelp = "(x) The angle (in radian) whose sine is x.", Position = 126)]
        public static double? Asin(object d) { if (d == null) return null; return System.Math.Asin(Convert.ToDouble(d)); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(r) The cosine of angle r (in radian). r = degrees*Pi/180.",
            ShortHelp = "(r) The cosine of angle r (in radian). r = degrees*Pi/180.", Position = 127)]
        public static double? Cos(object d) { if (d == null) return null; return System.Math.Cos(Convert.ToDouble(d)); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(x) The angle (in radian) whose cosine is x.",
            ShortHelp = "(x) The angle (in radian) whose cosine is x.", Position = 128)]
        public static double? Acos(object d) { if (d == null) return null; return System.Math.Acos(Convert.ToDouble(d)); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(r) The tangent of angle r (in radian). r = degrees*Pi/180.",
            ShortHelp = "(r) The tangent of angle r (in radian).", Position = 129)]
        public static double? Tan(object d) { if (d == null) return null; return System.Math.Tan(Convert.ToDouble(d)); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(x) The angle (in radian) whose tangent is x.",
            ShortHelp = "(x) The angle (in radian) whose tangent is x.", Position = 130)]
        public static double? Atn(object d) { if (d == null) return null; return System.Math.Atan(Convert.ToDouble(d)); }

        static Random r = new Random(); // Generates a series of random numbers from start of the run

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "( ) A random double-precision, floating number n (n >= 0 and n < 1).",
            ShortHelp = "( ) A random double-precision, floating number n (n >= 0 and n < 1).", Position = 135)]
        public static double Rnd() { return r.NextDouble(); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, 
            LongHelp = "(n, m) A random integer between n and m-1.",
            ShortHelp = "(n, m) A random integer between n and m-1.", Position = 135)]
        public static int? RndInt(object min, object max) { if (min == null || max==null) return null; return r.Next(Convert.ToInt32(min), Convert.ToInt32(max)); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(n) The absolute value of n.",
            ShortHelp = "(n) The absolute value of n.", Position = 140)]
        public static object Abs(object o) {
            if (o == null) return null; 
            if (Type.GetTypeCode(o.GetType()) == TypeCode.Double) return System.Math.Abs(Convert.ToDouble(o));
            return System.Math.Abs(Convert.ToInt32(o));
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(n) A string with the hexadecimal value of n.",
            ShortHelp = "(n) A string with the hexadecimal value of n.", Position = 141)]
        public static string Hex(object i) { if (i == null) return null; return Convert.ToInt32(i).ToString("x"); }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, SQL = true,
            LongHelp = "(n) A string with the octal value of n.",
            ShortHelp = "(n) A string with the octal value of n.", Position = 141)]
        public static string Oct(object n) { if (n == null) return null; return Convert.ToString(Convert.ToInt32(n), 8); }

        private static void UpdateTime() {
            if (VisEngine.simulatedTime)
                VisEngine.simulatedDateTime = VisEngine.simulatedDateTime.AddDays((double)VisEngine.simulatedTimeIncrement);
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement, Position = -20,
            ShortHelp = "( ) Recalculate the properties in all open forms and update the screen.\nRequery only where the rows formula has changed.")]
        public static bool Refresh() { // RefreshMode is ValueChange or higher.
            UpdateTime();
            VisEngine.MasterRefresh(); return true;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement, Position = -20,
            ShortHelp = @"(c) Recalculate the properties in component c's form and update the screen. Requery only where the rows formula has changed. 
c may be a form name. Then refresh all clones of the form. False when c is not an open form.")]
        public static bool RefreshForm(object frm) {
            return RefreshForm(frm, VisEngine.RefreshMode.ValueChange);
        }

        static bool RefreshForm(object frm, VisEngine.RefreshMode mode) {
            UpdateTime();
            if (frm == null) return false;
            if (VisEngine.refreshMode < mode) VisEngine.refreshMode = mode;
            VisComponent comp = frm as VisComponent;
            if (comp != null) {
                VisEngine.FormRefresh(comp.tpl.FormTpl);
                return true;
            }
            string frmName = frm as string;
            List<FormTemplate> clones = null;
            if (frmName != null && VisEngine.Forms.TryGetValue(frmName, out clones)) {
                for (int i = 0; i < clones.Count; i++)
                    if (clones[i] != null && clones[i].open) {
                        if (VisEngine.refreshMode < mode) VisEngine.refreshMode = mode;
                        VisEngine.FormRefresh(clones[i]);
                    }
                return true;
            }
            return false; // 
            //Token tk = VisEngine.curExprToken;
            //ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
            //    frmName + " must be a component or the name of an open form", tk);
            //throw new LoggedException();
        }
        
        [VisSystemAtt(VisSystemAtt.ItemType.Statement, Position = -20, 
            ShortHelp = @"( ) Requery the rows in all open forms. Recalculate the properties and update the screen.")]
        public static bool Requery() {
            UpdateTime();
            VisEngine.refreshMode = VisEngine.RefreshMode.Requery;
            VisEngine.MasterRefresh(); return true;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement, Position = -20,
            ShortHelp = "(c) Requery the rows in component c's form. Recalculate the properties in it and update the screen.\nFalse when c is not in an open form.")]
        public static bool RequeryForm(object frm) {
            return RefreshForm(frm, VisEngine.RefreshMode.Requery);
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement, Position = -19,
            ShortHelp = "(s, a, b ...) Open form s (\"name\" or Forms! name) and pass a, b ... to it as Param[0], Param[1] ... ",
            LongHelp = @"(s, a, b ...) Open form s (""name"" or Forms! name) and pass a, b ... to it as Param[0], Param[1] ... 
Result: The cloneID of the form. Close other open clones of s.")]
        public static int OpenForm(object form, Expression[] frmParams) {  // Allows form by name and as Forms! frmName 
            string frmName = form as string; // E.g. "frmName" or Forms! frmName with a closed Form.
            VisComponent frm = form as VisComponent; // Happens when Forms! frmName is an open form.
            if (frm != null) frmName = frm.Form().Name;
            if (frmName == null) return -1;
            int cloneID = 0;
            FormTemplate frmTpl = VisEngine.GetFormTemplate(frmName, ref cloneID);
            if (cloneID < 0) { // The form doesn't exist.
                Token t = VisEngine.curExprToken;
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                    "Form " + frmName + " doesn't exist as a file", t, 0);
                return -1;
            }
            VisEngine.CloseAllFormClones(frmName);
            return OpenMulti(frmName, frmParams);
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement, Position = -18,
            ShortHelp = "(s, a, b ...) Open form s (string name) and pass a, b ... to it as Param[0], Param[1] ... Can open multiple clones of the Form.",
            LongHelp = @"(s, a, b ...) Open form s (string name) and pass a, b ... to it as Param[0], Param[1] ...
Result: the cloneID of the form. If form s is open already, opens a new clone of it.")]
        public static int OpenMulti(object form, Expression[] parameters) {  // Should also allow form name without quote, i.e. path state Form ??
            string frmName = form as string;
            VisComponent frm = form as VisComponent;
            if (frm != null) frmName = frm.Form().Name;
            if (frmName == null) return -1;

            int cloneID;
            FormTemplate frmTpl = VisEngine.OpenForm(frmName, parameters, out cloneID);
            return cloneID;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement, Position = -17,
            ShortHelp = "(c) Close component c's Form. Usually c is the form. Return false if the form is still open, for instance because the user canceled the close.")]
        public static bool CloseForm(VisComponent cmp) {  // Closes the Form bundle of the component
            if (cmp == null) return true; // When closing a clone, it is cumbersome to the designer to check whether the clone slot is null. 
            (cmp.Form().GuiInstance as Form).Close();
            return ! Handled;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement, Position = -16,
            ShortHelp = "(s) Close all clones of Form s, e.g. CloseMulti(\"list1\"). Return false if some of the clones are still open, for instance because the user canceled the close.")]
        public static bool CloseMulti(string frmName) {  // Closes all clones of the component's form.
            if (frmName == null) return true; 
            VisEngine.CloseAllFormClones(frmName); 
            return !Handled;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement,
            LongHelp = "(c) Create a row in component c's primary table in memory. Refresh() shows the related visual components.\nUse Commit() or CommitOne(c) to save the changes in the database.",
            ShortHelp = "(c) Create a row in component c's primary table in memory.", Position = -15)]
        public static bool CreateRow(object comp) { // comp may be a component or a template
            Template tpl = comp as Template;
            if (comp is VisComponent) {
                //bundle = (comp as VisComponent).Bundle;
                tpl = ((VisComponent)comp).tpl;
            }
            if (tpl == null || tpl.RowTpl.RowsExpr == null) { // || tpl.Rows.primaryTable == null) {
                Token t = VisEngine.curExprToken;
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                    tpl == null ? "Expects a template" : "The template has no Rows", t, 0);
                return false;
            }
            Bundle bundle = tpl.RowBundle;

            System.Data.DataRow row = tpl.RowTpl.RowsExpr.RecordSet.NewRow();
            tpl.RowTpl.addedRows.Add(row);
            tpl.RowTpl.RowsExpr.RecordSet.Rows.Add(row);

            ConstructorInfo ciCreateCtlInst = tpl.Info.Type.GetConstructor(
            new System.Type[] { typeof(int), typeof(System.Data.DataRow), typeof(Template) });
            VisComponent viscomp = ciCreateCtlInst.Invoke(new object[] { 0, row, tpl }) as VisComponent;
            if (viscomp.GuiInstance != null) viscomp.GuiInstance.Tag = viscomp;
            CmpProp cmpProp = viscomp.propertyValues[VisComponent.RowsID];
            cmpProp.Value = new BundleRef(bundle, bundle.Instances.Count, null); 
            bundle.Instances.Add(cmpProp);
            // The template isn't dirty until something has been stored in the emtpy, new row. 

            //string targetName = tpl.FormTpl.Name + "_" + tpl.FormTpl.cloneID.ToString();
            //if (!VisEngine.DirtyDataTemplates.ContainsKey(targetName))
            //    VisEngine.DirtyDataTemplates.Add(targetName, tpl.FormTpl);
            //return true;
            //}

            return true;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement,
            LongHelp = "(c) Delete component c's row from the primary table in memory.\nRefresh() deletes the corresponding visual components.\nUse Commit() to save the changes in the database.",
            ShortHelp = "(c) Delete component c's row from the primary table in memory.", Position = -14)]
        public static bool DeleteRow(object comp) {
            VisComponent component = comp as VisComponent;
            if (component == null) return false;
            try {
                Template tpl = component.tpl;
                RecordSetExpr rows = tpl.RowTpl.RowsExpr;
                if (component.Row == null || rows == null || rows.primaryTable == null) 
                    throw(new Exception("The row cannot be deleted because it has no primary table"));
                if (component.Row.RowState == System.Data.DataRowState.Deleted) {
                    Debug.Print("Trying to delete the row twice: " + tpl.Name);
                    return false;
                }
                string error = VisEngine.DataMaps[0].getTable(tpl.RowTpl.RowsExpr.primaryTable).
                    PrepareDeleteCommand(component.Row, tpl);
                if (error !=null) throw(new Exception(error));
                tpl.RowTpl.addedRows.Remove(component.Row);
                component.Row.Delete();
                FormTemplate targetTpl = component.tpl.FormTpl;
                if (!targetTpl.DirtyDataTemplates.ContainsKey(tpl.Name))
                    targetTpl.DirtyDataTemplates.Add(tpl.Name, tpl);
            }
            catch (Exception e) {
                Token t = VisEngine.curExprToken;
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, e.Message, t, component.Index());
                throw new LoggedException();
            }
            component.Dispose();
            component.tpl.RowBundle.Instances.Remove(component.propertyValues[VisComponent.RowsID]);
            return true;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement,
            LongHelp = @"(c) Update the database according to changes in c's Form. Only CreateRow, DeleteRow and field changes
in primary tables are committed. CommitForm does nothing in simulation mode.",
            ShortHelp = "(c) Update the database according to changes in primary tables of c's Form. CommitForm does nothing in simulation mode.", Position = -13)]
        public static bool CommitForm(object comp) { // comp may be a component or a template
            Template tpl = comp as Template;
            if (tpl == null) tpl = (comp as VisComponent).tpl;
            FormTemplate frmTpl = tpl.FormTpl;
            foreach (var item in frmTpl.DirtyDataTemplates) { // Look at all templates in the form.
                CommitTpl(item.Value); 
            }
            frmTpl.DirtyDataTemplates.Clear();
            return true;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement,
            LongHelp = @"(c) Update the database according to changes in c's template. Only CreateRow, DeleteRow and field changes
in primary tables are committed. Next, retrieve automatic key fields and default values given by the database. Commit does nothing in simulation mode.",
            ShortHelp = "(c) Update the database according to changes in c's primary table. Next, retrieve automatic key fields and default values.", Position = -13)]
        public static bool CommitOne(object comp) { // The expression "tpl" may be a template or a component with index 0.
            Template tpl = comp as Template;
            if (tpl == null) tpl = (comp as VisComponent).tpl;
            // Should start a transaction here and roll it back if something goes wrong.
            CommitTpl(tpl);
            FormTemplate frmTpl = tpl.FormTpl;
            frmTpl.DirtyDataTemplates.Remove(tpl.Name);
            // Should commit the transaction here.
            return true;
        }

        private static void CommitTpl(Template tpl) {
            if (VisEngine.simulatedUpdate) return;
            if (tpl.RowTpl.RowsExpr == null) return;
            //foreach (Join join in tpl.Rows.joinList) {
            //    Table table = join.table;  
            Table table = VisEngine.DataMaps[0].getTable(tpl.RowTpl.RowsExpr.primaryTable);

            // Old solution: System.Data.DataTable addedTable = tpl.Rows.RecordSet.GetChanges(System.Data.DataRowState.Added);
            // Note: addedTable is a clone of the changed rows. UVis needs to change the rows that haven't got their autokey yet. SL 04-02-14.
            // A DataSet.Merge is needed to push them back into the Recordset. May not work with rows that haven't yet got their autokey.
            // The new solution is a uVis list of added rows in tpl.

            System.Data.DataTable modifiedTable = tpl.RowTpl.RowsExpr.RecordSet.GetChanges(System.Data.DataRowState.Modified);
            // modifiedTable is a clone too. UVis doesn't need to change these rows. It has been done directly in the RecordSet.
            // UVis only needs to generate SQL statements to update the database.
            // When the kernel had loaded the recordset, it used to set the last field to refer to the component. This was recorded as a change, and all rows were written back to the database.
            // The kernel now omits these component references except for recordsets that are targets for -=. As a consequence, they cannot be updated.

            if (table != null) {
                while (tpl.RowTpl.addedRows.Count > 0) { // FireInsert removes the row from addedRows.
                    System.Data.DataRow addedRow = tpl.RowTpl.addedRows[0];
                    tpl.RowTpl.addedRows.Remove(addedRow); // Remove add row indication whether creation succeeds or not.
                    table.FireInsertCommand(addedRow, tpl);
                }
                if (modifiedTable != null)
                    foreach (System.Data.DataRow modifiedRow in modifiedTable.Rows)
                        table.FireUpdateCommand(modifiedRow, tpl, table.Name); // join.prefix);
            }

            table.FireDeleteCommands(tpl);
            tpl.RowTpl.RowsExpr.RecordSet.AcceptChanges();
        }

        [VisSystemAtt(VisSystemAtt.ItemType.SystemMethod, Position = -12,
            ShortHelp = "(c) True if database data in c's Form have been changed, but not yet committed to the database.",
            LongHelp = @"(c) True if database data in c's Form have been changed, but not yet committed to the database.
A created row doesn't make the database dirty until something is stored in the row.")]
        public static bool IsDirty(object comp ) {  // The expression "tpl" may be a template or a component with index 0.
            Template tpl = comp as Template;
            if (tpl == null) tpl = (comp as VisComponent).tpl;
            FormTemplate frmTpl = tpl.FormTpl;
            bool result = frmTpl.DirtyDataTemplates.Count > 0;
            return result;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement,
            ShortHelp = "(c) Cancel the changes in c's rows that have not yet been committed to the database.\nIf c is a Form, cancel changes in all its templates.", Position = -11)]
        public static bool CancelUpdates(object comp) {
            Template tpl = comp as Template;
            if (tpl == null) tpl = (comp as VisComponent).tpl;
            FormTemplate frmTpl = tpl.FormTpl;
            if (frmTpl == tpl)
                frmTpl.DirtyDataTemplates.Clear();
            else 
                frmTpl.DirtyDataTemplates.Remove(tpl.Name);
            return true;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement, 
            LongHelp = "(c) Give component c keyboard focus, so that it receives all keyboard events.",
            ShortHelp = "(c) Give component c keyboard focus, so that it receives all keyboard events.", Position = -10)]
        public static bool SetFocus(VisComponent component) {
            if (component == null) return false;
            VisEngine.SelectedComponent = component;
            if (component.GuiInstance != null)
                ((Control)component.GuiInstance).Focus();
            else
                component.SetKeyFocus();
            return true;
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement,
            LongHelp = @"(s, i) Show string s and buttons i as a message box. Return user reply as a string, e.g. ""No"".
i=0: OK, i=1: OKCancel, i=2: YesNo, i=3: YesNoCancel, i=4: RetryCancel, b=5: AbortRetryCancel.",
            ShortHelp = "(s, i) Show string s and buttons i as a message box. Return user reply as a string, e.g. \"No\".", Position = -8)]
        public static string MsgBox(string msg, int bcode) {
            DialogResult res = DialogResult.None;
            MessageBoxButtons buttons;
            switch (bcode) {
            default: buttons = MessageBoxButtons.OK; break;
            case 1: buttons = MessageBoxButtons.OKCancel; break;
            case 2: buttons = MessageBoxButtons.YesNo; break;
            case 3: buttons = MessageBoxButtons.YesNoCancel; break;
            case 4: buttons = MessageBoxButtons.RetryCancel; break;
            case 5: buttons = MessageBoxButtons.AbortRetryIgnore; break;
            }
            res = System.Windows.Forms.MessageBox.Show(msg, "uVis", buttons);
            return res.ToString();
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement,
        ShortHelp = "(s, c) Show string s as a popup box next to component c", Position = -6)]
        public static void ShowPopup(string text, VisComponent source) {
            if (source == null) return;
            if (Popup != null) Popup.Dispose();
            PopupSource = source;
            Popup = new PopupForm(text, source);
            Popup.Show();
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement,
        ShortHelp = "(c) Close the popup box next to component c", Position = -5)]
        public static void ClosePopup(VisComponent source) {
            if (PopupSource == source) Popup.Dispose();
        }

        [VisSystemAtt(VisSystemAtt.ItemType.Statement,
            LongHelp = "(s) Log string s as a message in ErrorLog.txt",
            ShortHelp = "(s) Log string s as a message in ErrorLog.txt", Position = -3)]
        public static bool MsgLog(string msg) {
            Trace.WriteLine(msg); return true;
        }
    }
}
