﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Microsoft.Win32;

using System.IO;
using System.Text.RegularExpressions;

using System.Runtime.InteropServices;

namespace VisTool.UserConsole
{
    public partial class ConsoleForm : Form
    {
        [System.Runtime.InteropServices.DllImport("User32.dll")]
        private static extern bool SetForegroundWindow(System.IntPtr hWnd);

        public static System.Threading.EventWaitHandle ProgramStarted;

        public const string GlobalRegistry =  @"HKEY_CURRENT_USER\VisTool\";
        public const string RegistryGlobalArgs = "GlobalARGs";

        static string WriteGlobalArg(string[] args)
        {
            string s = "";
            if (args != null)
            {
                foreach (string arg in args)
                {
                    s += arg + "$";
                }
            }
            return s;
        }
        string[] ReadGlobalArg(string arg)
        {
            if (string.IsNullOrEmpty(arg)) return new string[] { };
            return arg.Split(new char[] { '$' }, StringSplitOptions.RemoveEmptyEntries);
        }
        
        [STAThread]
        static void Main(string[] args)
        {
            //ShareMem MemDB = new ShareMem();
            //if (MemDB.Init("VisToolArgs", 10240) != 0)
            //{
            //    MessageBox.Show("Error memory sharing");
            //}
            if (Registry.CurrentUser.OpenSubKey("VisTool") == null)
                Registry.CurrentUser.CreateSubKey("VisTool", RegistryKeyPermissionCheck.ReadWriteSubTree);


            Registry.SetValue(GlobalRegistry, RegistryGlobalArgs, WriteGlobalArg(args));
            
            bool createNew;
            ProgramStarted = new System.Threading.EventWaitHandle(false, System.Threading.EventResetMode.AutoReset, "VisToolStartEvent", out createNew);
            if (!createNew)
            {
                ProgramStarted.Set();
                System.Threading.Thread.Sleep(200);

                foreach (System.Diagnostics.Process p in System.Diagnostics.Process.GetProcessesByName(System.Diagnostics.Process.GetCurrentProcess().ProcessName))
                {
                    if (p.MainWindowHandle != IntPtr.Zero)
                    {
                        SetForegroundWindow(p.MainWindowHandle);
                    }
                }
                // exit the second thread
                return;
            }

            System.Windows.Forms.Application.SetCompatibleTextRenderingDefault(true);
            System.Windows.Forms.Application.EnableVisualStyles();

            System.Windows.Forms.Application.Run(new ConsoleForm());


        }

        //static Regex r = new Regex(@"^[\w\d_]+\.vis[m]*$", RegexOptions.Compiled | RegexOptions.IgnoreCase);
        public void RunVisEngine() {
            this.args = ReadGlobalArg(Registry.GetValue(GlobalRegistry, RegistryGlobalArgs, "").ToString());
            string mapPath = null;
            //if a user clicks a *.vism file
            if (args != null && args.Length > 0)  {
                mapPath = System.IO.Path.GetFullPath(args[0].Trim());
                if (!System.IO.File.Exists(mapPath))  {
                    MessageBox.Show(mapPath + " not found", "Studio", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                    OpenFileDialog dlg = new OpenFileDialog();
                    dlg.Filter = "VisTool DataMap File|*.vism";
                    dlg.Title = "Please select a datamap";
                    if (dlg.ShowDialog() == System.Windows.Forms.DialogResult.OK) {
                        mapPath = dlg.FileName;
                    }
                    else {
                        Application.Exit();
                    }
                }              
            }
            else // if no vism file given, pop up a dialog and let the user select a vism file
            {
                OpenFileDialog dlg = new OpenFileDialog();
                FileInfo fi = new FileInfo(Application.ExecutablePath);
                dlg.InitialDirectory = fi.Directory.FullName;
                dlg.Multiselect = false;
                dlg.Filter = "VisTool DataMap File|*.vism";
                dlg.Title = "Please select a datamap file";
                if (dlg.ShowDialog() == System.Windows.Forms.DialogResult.OK) {
                    mapPath = dlg.FileName;
                }
                else {
                    MessageBox.Show("Datamap not specified", "VisTool.exe", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                    Application.Exit();
                }
            }
            this.txtDebug.Text = "";
            try {
                VisTool.Parser.OpenMapFile(mapPath);

                if (VisEngine.ApplicationDirectory == "") {
                    MessageBox.Show("Datamap file is not correctly parsed", "VisTool.exe", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                    this.showDebugWindowToolStripMenuItem.Checked = true;
                    this.txtDebug.Visible = true;
                    this.Show();
                    return;
                }
            }
            catch (System.Exception ex) {
                MessageBox.Show(ex.ToString(), "VisTool.exe", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                this.showDebugWindowToolStripMenuItem.Checked = true;
                this.txtDebug.Visible = true;
                this.Show();
                return;
            }
            int bundleID;
            VisEngine.OpenForm(VisEngine.startUpForm, null, out bundleID);

        }





        /// <summary>
        /// Taskbar Icon
        /// </summary>
        System.Windows.Forms.NotifyIcon nIcon; 
        System.Windows.Forms.ContextMenu cm;
        public string[] args;
        
        void OnProgramStarted(object state, bool timeout)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new System.Threading.WaitOrTimerCallback(OnProgramStarted), state, timeout);
            }
            else
            {
                this.RunVisEngine();
            }
        }


        MenuItem itemRefresh;
        MenuItem itemClose;

        public ConsoleForm()
        {
            InitializeComponent();

            System.Threading.ThreadPool.RegisterWaitForSingleObject(ProgramStarted, OnProgramStarted, null, -1, false);
            Console.SetOut(new TextBoxWriter(this.txtDebug));

            bool b;
            try//restore menu options from Registry
            {
                b = bool.Parse(Registry.GetValue(GlobalRegistry,
                            prefixRegistryMenu + this.showDebugWindowToolStripMenuItem.Name, false).ToString());
                this.txtDebug.Visible = this.showDebugWindowToolStripMenuItem.Checked = b;

                if (!this.showDebugWindowToolStripMenuItem.Checked)
                {
                    this.WindowState = FormWindowState.Minimized;
                    this.ShowInTaskbar = false;
                }

            }
            catch { }

            try
            {
                string strSize = "";
                strSize = Registry.GetValue(GlobalRegistry, RegistrySize, "").ToString();
                if (!String.IsNullOrEmpty(strSize) && Regex.IsMatch(strSize, @"\d+,\d+"))
                {
                    int width = int.Parse(strSize.Split(new char[] { ',' })[0]);
                    int height = int.Parse(strSize.Split(new char[] { ',' })[1]);
                    this.Size = new Size(width, height);
                }
            }
            catch{}


            //-----------notification icon------------
            nIcon = new NotifyIcon();
            nIcon.Visible = true;
            try
            {
                System.IO.Stream stream = System.Reflection.Assembly.GetExecutingAssembly().GetManifestResourceStream("VisTool.VisIcon.ico");
                nIcon.Icon = new System.Drawing.Icon(stream);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
            this.nIcon.DoubleClick += new EventHandler(delegate
            {
                this.ShowInTaskbar = true;
                this.Show();
            });

            cm = new System.Windows.Forms.ContextMenu();
            itemClose = new MenuItem();
            itemClose.Click += new EventHandler(delegate { this.Close(); });
            cm.MenuItems.Add(itemClose);

            itemRefresh = new MenuItem();
            itemRefresh.Click += new EventHandler(delegate {
                this.hotKeyRefresh.UnregisterHotkey();
                this.hotKeyClose.UnregisterHotkey();
                RegisterCloseKey(Registry.GetValue(GlobalRegistry, RegistryCloseKey, Keys.F4).ToString());
                RegisterRefreshKey(Registry.GetValue(GlobalRegistry, RegistryRefreshKey, Keys.F9).ToString());

                this.RunVisEngine(); 
            });
            cm.MenuItems.Add(itemRefresh);
            
            this.nIcon.ContextMenu = cm;
            //-----------notification icon ends------------


            hotKeyRefresh = new VisTool.SystemHotKey();
            hotKeyClose = new VisTool.SystemHotKey();
            hotKeyRefresh.OnHotkey += new VisTool.HotkeyEventHandler(delegate { btnRefresh_Click(null, null); });
                        hotKeyClose.OnHotkey += new VisTool.HotkeyEventHandler(delegate
            {
                this.Close();
            });


            //-----------Register global keys-------------
            RegisterRefreshKey(Registry.GetValue(GlobalRegistry, RegistryRefreshKey, Keys.F9).ToString());
            RegisterCloseKey(Registry.GetValue(GlobalRegistry, RegistryCloseKey, Keys.F4).ToString());



        }



        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);
            this.RunVisEngine();
        }


        void hotKey_OnHotkey(int HotKeyID)
        {
            btnRefresh_Click(null, null);
        }

        private void btnRefresh_Click(object sender, EventArgs e)
        {
            this.RunVisEngine();
        }


        protected override void OnDeactivate(EventArgs e)
        {
            base.OnDeactivate(e);
            if (!this.showDebugWindowToolStripMenuItem.Checked)
            {
                this.Hide();
                this.ShowInTaskbar = false;
            }
        }

   

        protected override void OnClosed(EventArgs e)
        {
            base.OnClosed(e);
            this.hotKeyClose.UnregisterHotkey();
            nIcon.Dispose();
            nIcon = null;


            Registry.SetValue(GlobalRegistry, 
                prefixRegistryMenu + this.showDebugWindowToolStripMenuItem.Name, 
                this.showDebugWindowToolStripMenuItem.Checked);

        }
        private void showDebugWindowToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ToolStripMenuItem item = sender as ToolStripMenuItem;
            item.Checked = !item.Checked;
            this.txtDebug.Visible = item.Checked;
        }
        private void ConsoleForm_ResizeEnd(object sender, EventArgs e)
        {
            Registry.SetValue(GlobalRegistry, RegistrySize, this.Size.Width.ToString() + "," + this.Size.Height.ToString());

        }


        #region Global Hotkeys
        public VisTool.SystemHotKey hotKeyRefresh;
        public VisTool.SystemHotKey hotKeyClose;

        private const string RegistrySize = "ConsoleSize";
        private const string prefixRegistryMenu = "menu_";
        private const string RegistryRefreshKey = "RefreshKey";
        private const string RegistryCloseKey = "CloseKey";


        //private const int hotkeyaction_F5 = 90; //Any unique number for this app - need a different number for each hotkey
        //private const int WM_HOTKEY = 0x312;
        //private const int WM_CREATE = 0x1;
        //private const int WM_DESTROY = 0x2;
        //private const int MOD_ALT = 0x1; //ALT
        //private const int MOD_CONTROL = 0x2; //CTRL
        //private const int MOD_SHIFT = 0x4; //SHIFT
        //private const int VK_HOTKEY_F5 = 112;
        //private const int VK_HOTKEY_F1 = 112; 

        

      

        void RegisterRefreshKey(string keyCode)
        {
            string key = keyCode;
            if (string.IsNullOrEmpty(key)) { key = "F9"; }

            #region oldversion
            hotKeyRefresh.RegisterHotkey((Keys)Enum.Parse(typeof(Keys), key), this.Handle, 0);
            #endregion

            //KEY_REFRESH = (int)((Keys)Enum.Parse(typeof(Keys), key));
            //RegKey(Handle, hotkey_REFRESH, 0, KEY_REFRESH);

            this.btnRefresh.Text = itemRefresh.Text = "Refresh (" + key + ")";
        }
        void RegisterCloseKey(string keyCode)
        {
            string key = keyCode;
            if (string.IsNullOrEmpty(key)) { key = "F4"; }

            hotKeyClose.RegisterHotkey((Keys)Enum.Parse(typeof(Keys), key), this.Handle, 0);


            //KEY_CLOSE = (int)((Keys)Enum.Parse(typeof(Keys), key));
            //RegKey(Handle, hotkey_CLOSE, 0, KEY_CLOSE); 


            this.itemClose.Text = "Close (" + key + ")";
        }

        private void hotkeyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string newKey = ResetGlobalKeys(RegistryRefreshKey, "F9");

            RegisterRefreshKey(newKey);
            RegisterCloseKey(Registry.GetValue(GlobalRegistry, RegistryCloseKey, Keys.F4).ToString());

            //KEY_REFRESH = (int)((Keys)Enum.Parse(typeof(Keys), newKey));
            //RegKey(Handle, hotkey_REFRESH, 0, KEY_REFRESH);
            //RegKey(Handle, hotkey_CLOSE, 0, KEY_CLOSE); 
        }
        private void closeHotkeyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string newKey = ResetGlobalKeys(RegistryCloseKey, "F4");

            RegisterRefreshKey(Registry.GetValue(GlobalRegistry, RegistryRefreshKey, Keys.F9).ToString());
            RegisterCloseKey(newKey);

            //KEY_CLOSE = (int)((Keys)Enum.Parse(typeof(Keys), newKey));
            //RegKey(Handle, hotkey_REFRESH, 0, KEY_REFRESH);
            //RegKey(Handle, hotkey_CLOSE, 0, KEY_CLOSE); 
        }

        string ResetGlobalKeys(string subRegistry, string defaultKeyCode)
        {
            hotKeyClose.UnregisterHotkey();
            hotKeyRefresh.UnregisterHotkey();

            //UnRegKey(Handle, hotkey_REFRESH);
            //UnRegKey(Handle, hotkey_CLOSE); 


            HotKeyMsg msg = new HotKeyMsg(this, subRegistry, defaultKeyCode);
            msg.ShowDialog();
            try
            {
                Registry.SetValue(GlobalRegistry,
                    subRegistry, 
                    msg.KeyInput);
            }
            catch { }

            return msg.KeyInput;



        }
        #endregion

        private void txtDebug_DoubleClick(object sender, EventArgs e)
        {
            this.txtDebug.Text = "";

        }





        //[System.Runtime.InteropServices.DllImport("user32.dll")]
        //private static extern int RegisterHotKey(IntPtr hwnd, int id, int fsModifiers, int vk);
        //[System.Runtime.InteropServices.DllImport("user32.dll")]
        //private static extern int UnregisterHotKey(IntPtr hwnd, int id);


        //const int hotkey_REFRESH = 90; //热键ID
        //const int hotkey_CLOSE = 91; //热键ID

        //private const int WM_HOTKEY = 0x312; //窗口消息-热键
        //private const int WM_CREATE = 0x1; //窗口消息-创建
        //private const int WM_DESTROY = 0x2; //窗口消息-销毁
        //private const int MOD_ALT = 0x1; //ALT
        //private const int MOD_CONTROL = 0x2; //CTRL
        //private const int MOD_SHIFT = 0x4; //SHIFT


        ////private const int VK_HOTKEY_N = 0x4E; //N
        ////private const int VK_HOTKEY_D = 0x44; //D
        ////private const int VK_HOTKEY_H = 0x48; //H

        //private  int KEY_REFRESH = 115;
        //private  int KEY_CLOSE = 117;

        ///// <summary>
        ///// 注册热键
        ///// </summary>
        ///// <param name="hwnd">窗口句柄</param>
        ///// <param name="hotKey_id">热键ID</param>
        ///// <param name="fsModifiers">组合键</param>
        ///// <param name="vk">热键</param>
        //private void RegKey(IntPtr hwnd, int hotKey_id, int fsModifiers, int vk)
        //{
        //    bool result;
        //    if (RegisterHotKey(hwnd, hotKey_id, fsModifiers, vk) == 0)
        //    {
        //        result = false;
        //    }
        //    else
        //    {
        //        result = true;
        //    }
        //    if (!result)
        //    {
        //        System.Windows.Forms.MessageBox.Show("Cannot register global key");
        //    }
        //}

        ///// <summary>
        ///// 注销热键
        ///// </summary>
        ///// <param name="hwnd">窗口句柄</param>
        ///// <param name="hotKey_id">热键ID</param>
        //private void UnRegKey(IntPtr hwnd, int hotKey_id)
        //{
        //    UnregisterHotKey(hwnd, hotKey_id);
        //}
        //protected override void WndProc(ref Message m)
        //{
        //    base.WndProc(ref m);
        //    switch (m.Msg)
        //    {
        //        case WM_HOTKEY: //窗口消息-热键
        //            switch (m.WParam.ToInt32())
        //            {
        //                case hotkey_REFRESH: //Refresh
        //                    btnRefresh_Click(null, null);
        //                    break;
        //                case hotkey_CLOSE: //Close
        //                    this.ShowInTaskbar = true;
        //                    this.Show();
        //                    this.Activate();
        //                    if (MessageBox.Show("Are you sure to exit VisTool Runtime?",
        //                        "Exit VisTool",
        //                        MessageBoxButtons.OKCancel,
        //                        MessageBoxIcon.Question) == System.Windows.Forms.DialogResult.OK)
        //                    {
        //                        this.Close();
        //                    }

        //                    break;
        //                default:
        //                    break;
        //            }
        //            break;
        //        //case WM_CREATE: 
        //        //    RegKey(Handle, hotkey_REFRESH, 0, KEY_REFRESH); 
        //        //    RegKey(Handle, hotkey_CLOSE, 0, KEY_CLOSE); 
        //        //    break;
        //        case WM_DESTROY: 
        //            UnRegKey(Handle, hotkey_REFRESH); 
        //            UnRegKey(Handle, hotkey_CLOSE); 
        //            break;
        //        default:
        //            break;
        //    }
        //}
      
        
    }




    public class TextBoxWriter : System.IO.TextWriter
    {
        /// <summary>
        /// Equivalent of Action in .NET 3.5
        /// </summary>
        delegate void VoidAction();

        TextBox txtEditor;
        internal TextBoxWriter(TextBox txt)
        {
            txtEditor = txt;
        }

        public override Encoding Encoding
        {
            get { return Encoding.UTF8; }
        }

        public override void Write(char c)
        {
            Write(c.ToString());
        }

        public override void Write(char[] buffer, int index, int count)
        {
            Write(new string(buffer, index, count));
        }

        public override void Write(string text)
        {
            txtEditor.Text += text;
        }
    }











    public class ShareMem
    {
        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        public static extern IntPtr SendMessage(IntPtr hWnd, int Msg, int wParam, IntPtr lParam);

        [DllImport("Kernel32.dll", CharSet = CharSet.Auto)]
        public static extern IntPtr CreateFileMapping(int hFile, IntPtr lpAttributes, uint flProtect, uint dwMaxSizeHi, uint dwMaxSizeLow, string lpName);

        [DllImport("Kernel32.dll", CharSet = CharSet.Auto)]
        public static extern IntPtr OpenFileMapping(int dwDesiredAccess, [MarshalAs(UnmanagedType.Bool)] bool bInheritHandle, string lpName);

        [DllImport("Kernel32.dll", CharSet = CharSet.Auto)]
        public static extern IntPtr MapViewOfFile(IntPtr hFileMapping, uint dwDesiredAccess, uint dwFileOffsetHigh, uint dwFileOffsetLow, uint dwNumberOfBytesToMap);

        [DllImport("Kernel32.dll", CharSet = CharSet.Auto)]
        public static extern bool UnmapViewOfFile(IntPtr pvBaseAddress);

        [DllImport("Kernel32.dll", CharSet = CharSet.Auto)]
        public static extern bool CloseHandle(IntPtr handle);

        [DllImport("kernel32", EntryPoint = "GetLastError")]
        public static extern int GetLastError();

        const int ERROR_ALREADY_EXISTS = 183;

        const int FILE_MAP_COPY = 0x0001;
        const int FILE_MAP_WRITE = 0x0002;
        const int FILE_MAP_READ = 0x0004;
        const int FILE_MAP_ALL_ACCESS = 0x0002 | 0x0004;

        const int PAGE_READONLY = 0x02;
        const int PAGE_READWRITE = 0x04;
        const int PAGE_WRITECOPY = 0x08;
        const int PAGE_EXECUTE = 0x10;
        const int PAGE_EXECUTE_READ = 0x20;
        const int PAGE_EXECUTE_READWRITE = 0x40;

        const int SEC_COMMIT = 0x8000000;
        const int SEC_IMAGE = 0x1000000;
        const int SEC_NOCACHE = 0x10000000;
        const int SEC_RESERVE = 0x4000000;

        const int INVALID_HANDLE_VALUE = -1;

        IntPtr m_hSharedMemoryFile = IntPtr.Zero;
        IntPtr m_pwData = IntPtr.Zero;
        bool m_bAlreadyExist = false;
        bool m_bInit = false;
        long m_MemSize = 0;

        public ShareMem()
        {
        }
        ~ShareMem()
        {
            Close();
        }

       

        public int Init(string strName, long lngSize)
        {
            if (lngSize <= 0 || lngSize > 0x00800000) lngSize = 0x00800000;
            m_MemSize = lngSize;
            if (strName.Length > 0)
            {
                //创建内存共享体(INVALID_HANDLE_VALUE)
                m_hSharedMemoryFile = CreateFileMapping(INVALID_HANDLE_VALUE, IntPtr.Zero, (uint)PAGE_READWRITE, 0, (uint)lngSize, strName);
                if (m_hSharedMemoryFile == IntPtr.Zero)
                {
                    m_bAlreadyExist = false;
                    m_bInit = false;
                    return 2; //创建共享体失败
                }
                else
                {
                    if (GetLastError() == ERROR_ALREADY_EXISTS)  //已经创建
                    {
                        m_bAlreadyExist = true;
                    }
                    else                                         //新创建
                    {
                        m_bAlreadyExist = false;
                    }
                }
                //---------------------------------------
                //创建内存映射
                m_pwData = MapViewOfFile(m_hSharedMemoryFile, FILE_MAP_WRITE, 0, 0, (uint)lngSize);
                if (m_pwData == IntPtr.Zero)
                {
                    m_bInit = false;
                    CloseHandle(m_hSharedMemoryFile);
                    return 3; //创建内存映射失败
                }
                else
                {
                    m_bInit = true;
                    if (m_bAlreadyExist == false)
                    {
                        //初始化
                    }
                }
                //----------------------------------------
            }
            else
            {
                return 1; //参数错误     
            }

            return 0;     //创建成功
        }
        /// <summary>
        /// 关闭共享内存
        /// </summary>
        public void Close()
        {
            if (m_bInit)
            {
                UnmapViewOfFile(m_pwData);
                CloseHandle(m_hSharedMemoryFile);
            }
        }

        /// <summary>
        /// 读数据
        /// </summary>
        /// <param name="bytData">数据</param>
        /// <param name="lngAddr">起始地址</param>
        /// <param name="lngSize">个数</param>
        /// <returns></returns>
        public int Read(ref byte[] bytData, int lngAddr, int lngSize)
        {
            if (lngAddr + lngSize > m_MemSize) return 2; //超出数据区
            if (m_bInit)
            {
                Marshal.Copy(m_pwData, bytData, lngAddr, lngSize);
            }
            else
            {
                return 1; //共享内存未初始化
            }
            return 0;     //读成功
        }

        /// <summary>
        /// 写数据
        /// </summary>
        /// <param name="bytData">数据</param>
        /// <param name="lngAddr">起始地址</param>
        /// <param name="lngSize">个数</param>
        /// <returns></returns>
        public int Write(byte[] bytData, int lngAddr, int lngSize)
        {
            if (lngAddr + lngSize > m_MemSize) return 2; //超出数据区
            if (m_bInit)
            {
                Marshal.Copy(bytData, lngAddr, m_pwData, lngSize);
            }
            else
            {
                return 1; //共享内存未初始化
            }
            return 0;     //写成功
        }
    }
}
