﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using VisTool;

namespace uVisStudio.VisDesigner {
    public class FormInteraction {
        private Controller controller = Controller.Instance;
        public const int MinRectangleSize = 50;

        public FormInteraction() {
            VisEngine.ShowForm += ShowForm;
        }

        public static void AllDragEnter(object sender, DragEventArgs e) {
            var componentInfo = (ComponentInfo)e.Data.GetData(typeof(ComponentInfo));
            if (componentInfo != null && HookState.IsDesignMode()) {
                e.Effect = DragDropEffects.Copy;
            }
        }

        public void ShowForm(object sender, DesignModeEventArgs e) {
            var form = e.form;
            //form.Visible = true; Counteracts that the Visible property actually is False.
            form.AllowDrop = true;

            form.ResizeBegin -= this.FormLocationChanged;
            form.ResizeBegin += this.FormLocationChanged;
            form.ResizeEnd -= this.FormLocationChanged;
            form.ResizeEnd += this.FormLocationChanged;
            form.LocationChanged -= this.FormLocationChanged;
            form.LocationChanged += this.FormLocationChanged;
            form.SizeChanged -= this.FormLocationChanged;
            form.SizeChanged += this.FormLocationChanged;
            form.DragDrop -= this.FormDragDrop;
            form.DragDrop += this.FormDragDrop;
            form.DragEnter -= AllDragEnter;
            form.DragEnter += AllDragEnter;
            var formInstance = (VisForm)form.Tag;
            form.FormClosing -= FormsClosing;
            form.FormClosing += new FormClosingEventHandler(FormsClosing); // FormsClosing;
            StudioBase.Explorer.UpdateExplorer(); // The form templates are not visible until the form is open SL 18-05-13 
        }

        private void FormsClosing(object sender, FormClosingEventArgs e) {
            Form form = (Form)sender;
            FormTemplate frmTpl = ((VisForm)form.Tag).Bundle.template.FormTpl;
            var selEnum = SelectionBox.selections.GetEnumerator();
            if (selEnum.MoveNext() && selEnum.Current.Value.form == form) // Selections are in the form being closed.
                SelectionBox.DeleteAll(); // Close the blue boxes where the selections are.
            int openClones = 0;
            foreach (var cl in frmTpl.clones) if (cl != null && cl.open) openClones++;
            if (openClones > 1) return;
            if (controller.SaveFile(frmTpl, true))
                e.Cancel = true; // User said Cancel the closing.
        }

        public void FormDragDrop(object sender, DragEventArgs e) {
            var form = (Form)sender;
            ContainerComponent container = form.Tag as ContainerComponent;
            var screenPoint = new Point(e.X, e.Y);
            var componentInfo = (ComponentInfo)e.Data.GetData(typeof(ComponentInfo));
            container = HookState.FindContainer(container, screenPoint);
            if (container != null) {
                Point clientPoint = container.GuiInstance.PointToClient(screenPoint);
                this.controller.TemplateDropped(componentInfo, container, clientPoint);
            }
        }

        public void FormLocationChanged(object sender, EventArgs e) { // Used also for SizeChanged, etc.
            Template frmTpl = null;
            Form form = (Form)sender;
            VisComponent cmp = form.Tag as VisComponent;
            if (cmp == null) return;
            foreach (var kv in SelectionBox.selections) {
                SelectionBox box = kv.Value;
                if (box.form == form) {
                    box.blueBox.Left = form.Left + box.relativeX;
                    box.blueBox.Top = form.Top + box.relativeY;
                }
            }

            frmTpl = cmp.Bundle.template;
            VisPropertyBox.StudioGrid grid = VisPropertyBox.PropertyBoxManager.FindStudioGrid(frmTpl);
            if (grid == null) return;
            VisPropertyBox.PropertyBoxTemplate boxTpl = grid.Grid;
            VisPropertyBox.Category cat = boxTpl.Categories[0]; // Common properties.
            foreach (VisPropertyBox.PropertyGridLine prop in cat.Properties) {
                if (prop.Property.TemplateProperty.Name == "Top") prop.instanceValue.Text = form.Top.ToString();
                if (prop.Property.TemplateProperty.Name == "Left") prop.instanceValue.Text = form.Left.ToString();
                if (prop.Property.TemplateProperty.Name == "Height") prop.instanceValue.Text = form.Height.ToString();
                if (prop.Property.TemplateProperty.Name == "Width") prop.instanceValue.Text = form.Width.ToString();
            }
            // Don't use UpdateAllValues() because formulas will be evaluated and it causes circular references when Top or Left is being set.
            //uVisStudio.VisPropertyBox.PropertyBoxManager.UpdateAllValues();
        }
    }
}
