﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Diagnostics;
using System.Threading;
using System.Windows;
using System.Windows.Forms;
using System.Windows.Input;
using System.Windows.Media;
using AvalonDock;
using VisTool;
using uVisStudio.VisDatamap;
using uVisStudio.VisDesigner;
using uVisStudio.VisErrors;
using uVisStudio.VisExplorer;
using uVisStudio.VisPropertyBox;
using Application = System.Windows.Application;
using MessageBox = System.Windows.MessageBox;

namespace uVisStudio {
    public static class StudioBase {
        public static StudioWindow Studio;
        public static uVisStudio.VisExplorer.VisExplorer Explorer;
        public static VisualDatamapUC datamap;
        public static bool formulaEditorFocus = false; // Allows Hook to handle Ctrl-Z, etc. when formulaEditor doesn't.
        public static Stopwatch sw1 = new Stopwatch(); // For measurements
    }

    //-------------------------------------------------------------
    public partial class StudioWindow : Window {
        //public static bool Refresh = true; // Whether to refresh automatically.
        public static bool SuggestHelp;
        public const int MaxPropertyGrids = 8;
        public System.Windows.Controls.Image undoImage;
        public System.Windows.Controls.Image undoImageLight;
        public System.Windows.Controls.Image redoImage;
        public System.Windows.Controls.Image redoImageLight;
        static List<StudioGrid> StudioGrids = new List<StudioGrid>();

        public PropertyBoxManager propertyBoxManager;
        private Controller controller = Controller.Instance;

        public StudioWindow() {
            //this.Title = Directory.GetParent(VisEngine.ProgramDirectory).Name; // This also initializes VisEngine (the kernel).
            StudioBase.Studio = this;
            var formInteraction = new FormInteraction();
            Hook.StartHook();
            this.InitializeComponent();
            // The following settings work fine for Windows 7. In Windows 10 they cause the Studio window to open briefly, then minimize.
            //this.Left = System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Width * 0.6;//Screen.PrimaryScreen.Bounds.Right - Width;
            //this.Width = System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Width * 0.4;
            //this.Top = 0;
            //this.Height = System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Height;
            this.Title = "uVis:  " + VisEngine.ProgramDirectory;

            this.Closing += StudioWindowClosing;
            this.dockManager.Loaded += StudioWindowLoaded;
            this.KeyUp += StudioWindowKeyUp;

            undoImage = new System.Windows.Controls.Image();
            undoImage.Source = Properties.Resources.undoImage.GetImage();
            undoImage.Height = 16;
            undoImage.Width = 16;

            undoImageLight = new System.Windows.Controls.Image();
            undoImageLight.Source = Properties.Resources.undoImageLight.GetImage();
            undoImageLight.Height = 16;
            undoImageLight.Width = 16;

            redoImage = new System.Windows.Controls.Image();
            redoImage.Source = Properties.Resources.redoImage.GetImage();
            redoImage.Height = 16;
            redoImage.Width = 16;

            redoImageLight = new System.Windows.Controls.Image();
            redoImageLight.Source = Properties.Resources.redoImageLight.GetImage();
            redoImageLight.Height = 16;
            redoImageLight.Width = 16;

            this.undoItem.Icon = undoImage;
            this.redoItem.Icon = redoImage;
            this.undoItem.Height = 16; this.undoItem.Width = 25;
            this.redoItem.Height = 16; this.redoItem.Width = 25;
            this.undoItem.MouseEnter += this.UndoItemMouseEnter;
            this.redoItem.MouseEnter += this.RedoItemMouseEnter;

            this.undoItem.MouseLeave += this.UndoItemMouseLeave;
            this.redoItem.MouseLeave += this.RedoItemMouseLeave;

            this.undoItem.Click += new RoutedEventHandler(CallUndoCommand);
            this.redoItem.Click += new RoutedEventHandler(CallRedoCommand);

            //this.undoImage.IsMouseDirectlyOverChanged += new DependencyPropertyChangedEventHandler(

            this.LimitedRowsCB.Click += LimitedRowsClick;
            this.MaxRows.LostFocus += MaxRowsLostFocus;
            this.Zoom.AcceptsReturn = false;
            this.Zoom.LostFocus += ZoomLostFocus;
            this.helpCB.Click += SuggestionClick;
            this.refreshCB.Click += AutoRefreshClick;
            this.simulatedTimeCB.Click += SimulatedTimeCBClick;
            this.simulatedDateTextbox.LostFocus += simulatedDateTextboxLostFocus;
            this.simulatedUpdateCB.Click += simulatedUpdateCBClick;
            this.dblBufCB.Click += dblBufCBClick;
            this.dayIncrementTextbox.LostFocus += dayIncrementTextboxLostFocus;
            //this.Closing += new System.ComponentModel.CancelEventHandler(StudioWindow_Closing);
            OpenFileCommand.MapFileOpened += MapFileOpened;
            //VisEngine.ShowForm += UpdateDesignSelection; Doesn't change as a result of Refresh()

            //create property grids
            StudioGrids = new List<StudioGrid>(MaxPropertyGrids);
            //StudioGrid grid1 = new StudioGrid(this.dockContent1, this.dock1); // create the first property grid defined in the xaml file
            this.dockContent1.Show(dockManager, AnchorStyle.Top);
            //grid1.Dock.Show();
            //StudioGrids.Add(grid1);
            for (int i = 0; i < MaxPropertyGrids; i++) {
                PropertyBoxTemplate pBoxTemplate = new PropertyBoxTemplate();
                pBoxTemplate.Name = "dock" + (i + 2);
                DockableContent dockContent = new DockableContent();
                dockContent.Content = pBoxTemplate;
                DockablePane dockPane = new DockablePane();
                this.reszingPanel.Children.Add(dockPane);
                dockPane.Items.Add(dockContent);
                StudioGrid sGrid = new StudioGrid(dockContent, pBoxTemplate);
                StudioGrids.Add(sGrid);
                if (i == 0) {
                    sGrid.Dock.Show();
                }
            }

            this.propertyBoxManager = new PropertyBoxManager(dockManager, StudioGrids);
        }

        public static void PushGridToTheFarRight(StudioGrid Grid) {
            if (!StudioGrids.Contains(Grid)) {
            }
        }

        public void CallRedoCommand(object sender, RoutedEventArgs e) {
            UndoManager.Redo();
        }

        public void CallUndoCommand(object sender, RoutedEventArgs e) {
            UndoManager.Undo();
        }

        public void ChangeUndoImage(object sender, DependencyPropertyChangedEventArgs e) {
            if (undoItem.IsMouseDirectlyOver)
                undoItem.Icon = undoImageLight;
            else
                undoItem.Icon = undoImage;
        }

        public void RedoItemMouseEnter(object sender, System.Windows.Input.MouseEventArgs e) {
            redoItem.Icon = redoImageLight;
        }

        public void RedoItemMouseLeave(object sender, System.Windows.Input.MouseEventArgs e) {
            redoItem.Icon = redoImage;
        }

        public void UndoItemMouseEnter(object sender, System.Windows.Input.MouseEventArgs e) {
            undoItem.Icon = undoImageLight;
        }

        public void UndoItemMouseLeave(object sender, System.Windows.Input.MouseEventArgs e) {
            undoItem.Icon = undoImage;
        }

        private void MapFileOpened(object sender, EventArgs e) {
            this.simulatedDateTextbox.Text = SystemItems.Now().ToString();
            this.LimitedRowsCB.IsChecked = VisEngine.LimitedRows;
            this.simulatedTimeCB.IsChecked = VisEngine.simulatedTime;
            this.simulatedUpdateCB.IsChecked = VisEngine.simulatedUpdate;
            this.dayIncrementTextbox.Text = VisEngine.simulatedTimeIncrement.ToString().Replace(",", ".");
            this.MaxRows.Text = VisEngine.MaxRows.ToString();
        }

        private void StudioWindowKeyUp(object sender, System.Windows.Input.KeyEventArgs e) {
            if (e.Key == Key.Z && Keyboard.Modifiers == ModifierKeys.Control) { // Ctrl-Z
                //Debug.Print("StudioWindow - Ctrl-Z");
                UndoManager.Undo();
            } else if (e.Key == Key.Y && Keyboard.Modifiers == ModifierKeys.Control) { // Ctrl-Y
                //Debug.Print("Studiowindow - Ctrl-Y");
                UndoManager.Redo();
            } else if (e.Key == Key.Delete && Keyboard.Modifiers == ModifierKeys.Control) { // Ctrl-Delete
                HookState.DeleteSelected();
            } else if (e.Key == Key.S && Keyboard.Modifiers == ModifierKeys.Control) { // Ctrl-S
                var studioGrid = PropertyBoxManager.SelectedStudioGrid;
                Template tpl;
                if (studioGrid != null && (tpl = studioGrid.Grid.Tpl) != null) {
                    this.controller.SaveFile(tpl.FormTpl, false);
                    return;
                }
                MessageBox.Show("No Form is selected", "Warning");
            } else if (e.Key == Key.S && Keyboard.Modifiers == (ModifierKeys.Control | ModifierKeys.Shift)) { // Ctrl-Shft-S
                this.controller.SaveAll(false);
            }
        }

        private void StudioWindowLoaded(object sender, RoutedEventArgs e) {
            string file = VisEngine.ProgramDirectory + "\\WindowPositions.xml";
            if (File.Exists(file)) {
                System.Globalization.CultureInfo c = Thread.CurrentThread.CurrentCulture;
                try {
                    Thread.CurrentThread.CurrentCulture = VisEngine.codeCulture; // Dock manager will use English number formats and separators.
                    dockManager.RestoreLayout(file);
                }
                catch (Exception ex) {
                    MessageBox.Show("Cannot read file " + file + "\nThe regional number settings may not match the settings assumed in the file.\n\n" + ex, "uVis Studio"); //, MessageBoxButtons.OK);
                }
                Thread.CurrentThread.CurrentCulture = c; // Restore cultural settings.
            }
        }

        private void StudioWindowClosing(object sender, System.ComponentModel.CancelEventArgs e) {
            this.controller.SaveAll(true);
            VisEngine.CloseAllForms(); // The last form should not close the database connection. The designer may want to open a vis-file.
            if (VisEngine.DataMaps.Count > 0 && VisEngine.DataMaps[0].Connection != null)
                VisEngine.DataMaps[0].Connection.Dispose();
            propertyBoxManager.CloseGrids();
            //var file = VisEngine.ProgramDirectory + "\\WindowPositions.xml";
            //MessageBoxResult res = MessageBox.Show("Save Studio layout?", "Save layout?", MessageBoxButton.YesNo);
            //if (res == MessageBoxResult.Yes) {
            //    System.Globalization.CultureInfo c = Thread.CurrentThread.CurrentCulture;
            //    Thread.CurrentThread.CurrentCulture = VisEngine.codeCulture; // Dock manager will use English number formats and separators.
            //    dockManager.SaveLayout(file);
            //    Thread.CurrentThread.CurrentCulture = c; // Restore cultural settings.
            //}
        }

        private void LimitedRowsClick(object sender, RoutedEventArgs e) {
            VisEngine.LimitedRows = (bool)LimitedRowsCB.IsChecked;
        }

        private void SimulatedTimeCBClick(object sender, RoutedEventArgs e) {
            VisEngine.simulatedTime = (bool)simulatedTimeCB.IsChecked;
        }

        private void MaxRowsLostFocus(object sender, RoutedEventArgs e) {
            int n;
            if (String.IsNullOrWhiteSpace(MaxRows.Text)) return;
            if (Int32.TryParse(MaxRows.Text.Trim(), out n)) {
                VisEngine.MaxRows = n;
            } else {
                MessageBox.Show(MaxRows.Text + " is not a number");
            }
        }

        private void ZoomLostFocus(object sender, RoutedEventArgs e) {
            if (String.IsNullOrWhiteSpace(Zoom.Text)) return;
            float zoom;
            float.TryParse(Zoom.Text.Trim(), System.Globalization.NumberStyles.AllowDecimalPoint, VisEngine.codeCulture, out zoom);
            if (zoom >0) {
                SystemItems.Zoom = zoom;
            } else {
                MessageBox.Show("This is not a zoom factor. Try e.g. 1.2");
            }
        }

        private void simulatedDateTextboxLostFocus(object sender, RoutedEventArgs e) {
            if (String.IsNullOrWhiteSpace(simulatedDateTextbox.Text)) return;
            DateTime d;
            if (DateTime.TryParse(simulatedDateTextbox.Text.Trim(), VisEngine.codeCulture, System.Globalization.DateTimeStyles.None, out d)) {
                VisEngine.simulatedDateTime = d;
            } else {
                MessageBox.Show(simulatedDateTextbox.Text + " is not a date");
                //simulatedDateTextbox.Text = VisEngine.simulatedDateTime.ToString();
            }
        }

        private void simulatedUpdateCBClick(object sender, RoutedEventArgs e) {
            VisEngine.simulatedUpdate = (bool)simulatedUpdateCB.IsChecked;
        }

        private void dblBufCBClick(object sender, RoutedEventArgs e) {
            VisEngine.doubleBuffer = (bool)dblBufCB.IsChecked;
        }

        private void dayIncrementTextboxLostFocus(object sender, RoutedEventArgs e) {
            double inc;
            try {
                inc = double.Parse(dayIncrementTextbox.Text.Trim(), new System.Globalization.CultureInfo("en-AU"));
                VisEngine.simulatedTimeIncrement = inc;
            }
            catch (Exception) {
                MessageBox.Show(dayIncrementTextbox.Text + " is not a number");
                dayIncrementTextbox.Text = VisEngine.simulatedTimeIncrement.ToString();
            }
        }

        private void SuggestionClick(object sender, RoutedEventArgs e) {
            if ((bool)helpCB.IsChecked) SuggestHelp = true;
            else SuggestHelp = false;
        }

        private void AutoRefreshClick(object sender, RoutedEventArgs e) {
            //VisEngine.CompileOnly = ! (bool)refreshCB.IsChecked;
            //if ((bool)refreshCB.IsChecked) Refresh = true;
            //else Refresh = false;
        }

        private void FileMenuItemClick(object sender, RoutedEventArgs e) {
            if (sender.Equals(saveCurrentMI)) {
                var studioGrid = PropertyBoxManager.SelectedStudioGrid;
                Template tpl;
                if (studioGrid != null && (tpl = studioGrid.Grid.Tpl) != null) {
                    this.controller.SaveFile(tpl.FormTpl, false);
                    return;
                }
                MessageBox.Show("No Form is selected", "Warning");
            }
            if (sender.Equals(saveAllMI)) {
                this.controller.SaveAll(false);
            }
            if (sender.Equals(saveStudioLayout)) {
                var file = VisEngine.ProgramDirectory + "\\WindowPositions.xml";
                System.Globalization.CultureInfo c = Thread.CurrentThread.CurrentCulture;
                Thread.CurrentThread.CurrentCulture = VisEngine.codeCulture; // Dock manager will use English number formats and separators.
                dockManager.SaveLayout(file);
                Thread.CurrentThread.CurrentCulture = c; // Restore cultural settings.
            }
            if (sender.Equals(exitMI)) {
                Application.Current.Shutdown();
            }
        }

        private void ViewMenuItemClick(object sender, RoutedEventArgs e) {
            if (sender.Equals(explorerMenuItem)) explorerContent.Show(dockManager);
            if (sender.Equals(toolboxMenuItem)) toolboxContent.Show(dockManager);
            if (sender.Equals(datamapMenuItem)) datamapContent.Show(dockManager);
            if (sender.Equals(errorListMenuItem)) errorListContent.Show(dockManager);
            if (sender.Equals(propertyGridMenuItem)) dockContent1.Show(dockManager);
            if (sender.Equals(defaultMenuItem)) {
                var file = VisEngine.ProgramDirectory + "\\DefaultWindowPositions.xml";
                if (File.Exists(file)) {
                    dockManager.RestoreLayout(file);
                }
            }
        }

        private void UpdateDesignSelection(object sender, EventArgs e) {
            if (this.interactionCombo.SelectedIndex == 0) {
                VisEngine.InteractionMode = VisEngine.interactionMode.AsEndUser;
                SelectionBox.DeleteAll();
            } else if (this.interactionCombo.SelectedIndex == 1) {
                VisEngine.InteractionMode = VisEngine.interactionMode.AsDesigner;
            } else if (this.interactionCombo.SelectedIndex == 2) {
                VisEngine.InteractionMode = VisEngine.interactionMode.DesignWithControl;
            } else if (this.interactionCombo.SelectedIndex == 3) {
                VisEngine.InteractionMode = VisEngine.interactionMode.DesignWithAlt;
            }
        }

        private void RefreshClick(object sender, RoutedEventArgs e) {
            VisEngine.refreshMode = VisEngine.RefreshMode.FormulaChange; // Recompile and requery.
            SystemItems.Refresh();
        }

        private void RequeryClick(object sender, RoutedEventArgs e) {
            SystemItems.Requery();
        }
    }
}
