﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using VisTool;

namespace uVisStudio.VisErrors {
    public partial class VisErrorList : UserControl {
        public VisErrorList() {
            InitializeComponent();
            ErrorLogging.Errors.CollectionChanged += ShowErrors;
            ErrorList.MouseDoubleClick += ErrorListMouseDoubleClick;
            ErrorList.KeyDown += ErrorListKeyDown;
        }

        private void ErrorListKeyDown(object sender, KeyEventArgs e) {
            if (e.Key != Key.Enter) return;
            if (this.ErrorList.SelectedItem == null) return;
            var errorItem = (ErrorItem)this.ErrorList.SelectedItem;
            VisPropertyBox.PropertyBoxManager.ShowError(errorItem.Error);
        }

        private void ErrorListMouseDoubleClick(object sender, MouseButtonEventArgs e) {
            if (this.ErrorList.SelectedItem == null) return;
            var errorItem = (ErrorItem)this.ErrorList.SelectedItem;
            VisPropertyBox.PropertyBoxManager.ShowError(errorItem.Error);
        }

        public void ShowErrors(object sender, NotifyCollectionChangedEventArgs e) {
            ErrorList.Items.Clear();
            var errors = ErrorLogging.Errors.Distinct();
            foreach (var error in errors) {
                string text = error.formName + ": " + error.templateName;
                if (error.mode == VisTool.ErrorMode.Vism)
                    text += " " + error.propertyName + ", line " + error.token.line + " " + error.token.val + ": " + error.ErrorMessage;
                else
                    text += (ErrorMode.RunTime == error.mode && error.Index >= 0 ? "[" + error.Index + "]" : "") +
                        "! " + error.propertyName + ": " + error.ErrorMessage;

                var errorItem = new ErrorItem(error, text, Brushes.Red);
                ErrorList.Items.Add(errorItem);
            }
        }
    }

    public class ErrorItem {
        public Error Error { get; set; }
        public Brush Color { get; set; }
        public string Message { get; set; }

        public ErrorItem(Error error, string message, Brush color) {
            this.Error = error;
            this.Message = message;
            this.Color = color;
        }
    }

    public class TemplateErrorEventArgs : EventArgs {
        public readonly Error Error;
        public readonly Template Template;
        public readonly TplProperty Property;
        public readonly VisComponent cmp;

        public TemplateErrorEventArgs(Error error, Template template, TplProperty property) {
            this.Error = error;
            this.Template = template;
            this.Property = property;

            var bundle = template.RowBundle.Instances;
            if (error.Index >= 0 && error.Index < template.RowBundle.Instances.Count) {
                cmp = bundle[error.Index].cmp;
            }
        }
    }
}
