﻿using System.IO;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using VisTool;

namespace uVisStudio.VisExplorer {
    public partial class RenameFileDialog : Window {
        private string fileNameWithExtension;
        private Controller controller = Controller.Instance;

        public RenameFileDialog() {
            this.InitializeComponent();
            this.SetupEvents();
            this.textBox.Focus();
            this.Background = ColorConstants.DialogWindowBackground();
        }

        public void SetFile(string file) {
            this.fileNameWithExtension = this.textBox.Text = Path.GetFileName(file);
        }

        private void SetupEvents() {
            this.textBox.LostFocus += TextBoxLostFocus;
            this.KeyDown += RenameFileDialogKeyDown;
        }

        private void Rename() {
            int cloneID = 0;
            FormTemplate frmTpl = null;
            string oldFrmName = null;
            if (this.fileNameWithExtension.EndsWith(".vis", System.StringComparison.InvariantCultureIgnoreCase)) {
                oldFrmName = this.fileNameWithExtension.Remove(this.fileNameWithExtension.Length - 4);
                frmTpl = VisEngine.GetFormTemplate(oldFrmName, ref cloneID);
            }
            string newFileName = this.textBox.Text.Trim();
            if (this.fileNameWithExtension == newFileName) return; // No change.
            if (frmTpl != null) {
                MessageBox.Show("Cannot rename the form here while it is open.\r\nEdit the form name in the property grid.");
                return;
            }
            if (newFileName.EndsWith(".vis", System.StringComparison.InvariantCultureIgnoreCase)) { // New must be a form name.
                string newFrmName = newFileName.Remove(newFileName.Length - 4).Trim(); // Without extension.
                if (!Parser.IsIdent(newFrmName)) {
                    MessageBox.Show(newFrmName + " is not an identifier");
                    return;
                }
                newFileName = newFrmName + ".vis";
            }
            this.controller.RenameFile(this.fileNameWithExtension, newFileName);
        }

        private void RenameFileDialogKeyDown(object sender, KeyEventArgs e) {
            if (e.Key == Key.Enter) {
                Rename();
                this.Close();
            }
            if (e.Key == Key.Escape) {
                this.Close();
            }
        }

        private void TextBoxLostFocus(object sender, RoutedEventArgs e) {
        }

        private void OkButtonClick(object sender, RoutedEventArgs e) {
            Rename();
            this.Close();
        }

        private void CancelButtonClick(object sender, RoutedEventArgs e) {
            this.Close();
        }
    }
}
