﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Reflection;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Linq;
using AvalonDock;
using VisTool;
using uVisStudio.VisErrors;
using uVisStudio.VisExplorer;
using uVisStudio.VisDesigner;

namespace uVisStudio.VisPropertyBox {
    public partial class PropertyBoxTemplate : UserControl {
        public Template Tpl;
        public VisComponent cmp;
        public PropertyGridLine SelectedProperty;
        public bool isSomeCopyProperties = false;
        public DateTime LastUsed = DateTime.Now;

        public List<PropertyGridLine> listSelectedProperties = new List<PropertyGridLine>();

        public List<Category> Categories { get; set; }

        public PropertyBoxTemplate() {
            this.Categories = new List<Category>();
            this.InitializeComponent();
            this.menu.PropertyGrid = this;
            this.menu.SetupEvents();
        }

        public void Reset() {
            this.Tpl = null;
            this.cmp = null;
            this.Categories.Clear();
            //this.menu.Reset(); // Lost in event handling. Never returns to VisForm.Close.
            this.Update();
        }

        public void ShowError(Error error) {
            foreach (var category in Categories) {
                foreach (var property in category.Properties) {
                    var propertyName = property.Property.TemplateProperty.Name;
                    if (propertyName.Equals(error.propertyName)) {
                        property.Select(error);
                    }
                }
            }
        }

        public void CreateProperties(Template template) {
            VisComponent instance = null;
            if (template.RowBundle != null && template.RowBundle.Instances.Count > 0) {
                this.cmp = template.RowBundle.Instances[0].cmp;
            } else
                this.cmp = null;
            this.Tpl = template;
            this.menu.VisComponent = instance;
            this.menu.CurrentTemplate = template;
            this.menu.SelectTemplate(template);

            CreateProperties(Tpl, cmp);
            this.Update();
        }

        public void CreateProperties(VisComponent instance) {
            var template = instance.tpl;
            this.Tpl = template;
            this.cmp = instance;
            this.menu.VisComponent = instance;
            this.menu.CurrentTemplate = template;
            this.menu.SelectTemplate(template);

            CreateProperties(Tpl, cmp);
            this.Update();
        }

        public void UpdateProperties() {
            if (cmp == null) {
                if (Tpl.RowBundle == null || Tpl.RowBundle.Instances.Count == 0) return;
                cmp = Tpl.RowBundle.Instances[0].cmp;
            }
            CreateProperties(Tpl, cmp);
            this.Update();
        }

        public void UpdateMenu() {
            if (Tpl != null)
                this.menu.SelectTemplate(Tpl);
        }

        public void deleteThisComponent(object sender, RoutedEventArgs e) {
            VisTool.Template tpl = PropertyBoxManager.SelectedStudioGrid.Grid.Tpl;
            if (tpl == null) return;
            //if (MessageBox.Show("Want to delete template " + tpl.Name + " ?", "", MessageBoxButton.OKCancel) == MessageBoxResult.OK) {
            SelectionBox.DeleteAll();
            Controller ctrl = Controller.Instance;
            if (ctrl == null) return;
            ctrl.DeleteTemplates(new List<Template>() { tpl });
        }

        public void selectThisComponent(object sender, RoutedEventArgs e) {
            VisComponent cmp = PropertyBoxManager.SelectedStudioGrid.Grid.cmp;
            VisTool.Template tpl = PropertyBoxManager.SelectedStudioGrid.Grid.Tpl;
            if (cmp == null || cmp.Disposed) { // Find another component if possible.
                cmp = null;
                    if (tpl.RowBundle != null && tpl.RowBundle.Instances.Count > 0) 
                        cmp = tpl.RowBundle.Instances[0].cmp;
            }
            if (cmp == null)
                MessageBox.Show("There is no component to show.\r\nThe property values are old; the formulas are okay.");
            else {
                SelectionBox.DeleteAll();
                SelectionBox sel = new SelectionBox(cmp);
                TemplateHandler.SelectTemplate(cmp, PropertyBoxManager.SelectedStudioGrid);
                sel.form.Focus();
            }
        }

        public void Update(object sender, RoutedEventArgs e) {
            this.Update();
        }

        public void Update() {
            var isCategorized = this.menu.categorized.IsChecked == true;
            if (isCategorized) {
                var controlItems = GetCategorizedItems();
                propertyBox.Update(controlItems);
            } else {
                var controlItems = GetAlphabeticalItems();
                propertyBox.Update(controlItems);
            }
        }

        private List<UserControl> GetCategorizedItems() {
            var controls = new List<UserControl>();
            foreach (var category in Categories) {
                controls.Add(category);
                if (!category.IsExpanded) continue;
                category.Properties.Sort((a, b) => (a.Property.attr.Position != b.Property.attr.Position) ?
                    a.Property.attr.Position.CompareTo(b.Property.attr.Position) :
                    a.Property.TemplateProperty.Name.CompareTo(b.Property.TemplateProperty.Name));   
                // Design properties don't have positions, so sort by name.
                foreach (var property in category.Properties) {
                    controls.Add(property);
                }                 
            }
            return controls;
        }

        public PropertyGridLine GetPropertyGUIByName(string name) {
            foreach (var category in Categories)
                foreach (PropertyGridLine property in category.Properties)
                    if (property.nameEditor.Text.Equals(name))
                        return property;
            //This should not happen.
            return null;
        }

        private List<UserControl> GetAlphabeticalItems() {
            var controls = new List<PropertyGridLine>();
            foreach (var category in Categories) {
                foreach (var property in category.Properties) {
                    controls.Add(property);
                }
            }
            controls.Sort((a, b) => a.Property.TemplateProperty.Name.CompareTo(b.Property.TemplateProperty.Name));
            return new List<UserControl>(controls);
        }

        public void CreateProperties(Template template, VisComponent instance) {
            var categoryMap = CreateCategories(template, instance);
            UpdateCategories(categoryMap);
        }

        private List<CategoryData> CreateCategories(Template template, VisComponent instance) { // Another CreateCategories is in PropertyBoxMenu
            var categories = new List<CategoryData>(10) // Insert the known categories in a designer-friendly sequence. SL 22-04-13
            { new CategoryData("Common"), new CategoryData("Special"), new CategoryData("Designer"), new CategoryData("EventHandler") };
            string categoryName;
            ComponentInfo compInfo = VisEngine.ComponentByName[template.Info.Type.Name];

            StudioProperty property;
            TplProperty t;
            // Add an artificial Type Property to category.Properties.
            t = new VisTool.TplProperty(template, "Type", null, VisTool.PropertyKind.Standard);
            t.Formula = template.Info.Type.Name;
            property = new TypeProperty(template, t, null);
            property.attr = new VisPropAtt(VisPropAtt.PropertyType.Property, null);
            property.attr.Position = 2;
            AddCategorizedItem(categories, "Common", property);

            // Look at all value properties with a formula and properties to be shown initially.
            foreach (TplPropValue tplProp in template.valPropArray) {
                if (tplProp.cmpPropIndex == VisComponent.NameID) { // Name property
                    property = new NameProperty(template, tplProp, instance);
                    //property.attr = new VisPropAtt(VisPropAtt.PropertyType.Property, " ");
                    //property.attr.Position = 0;
                } else if (tplProp.cmpPropIndex == VisComponent.RowsID) {
                    property = new RowsProperty(template, tplProp, instance);
                } else if (tplProp.Formula != null || tplProp.PropInfo.VisAttribute.ShowInitially) { // Show also formulas with " " property.
                    property = new StudioProperty(template, tplProp, instance);
                } else
                    continue; // Don't show the property.
                property.attr = tplProp.PropInfo.VisAttribute;
                categoryName = tplProp.PropInfo.VisAttribute.PropertyCategory;

                if (property.attr == null) { // Property without attribute will cause null-ref in Sort, etc. Give them a dummy attribute
                    Debug.Print(Tpl.Name + "! " + property.TemplateProperty.Name + " has no VisPropAtt");
                    property.attr = new VisPropAtt(VisPropAtt.PropertyType.Property, " ");
                    property.attr.Position = 0;
                }
                AddCategorizedItem(categories, categoryName, property);
            }

            // Look at all event properties with a formula or properties to be shown initially.
            foreach (TplPropEvent tplProp in template.evtPropArray) {
                if (tplProp.Formula != null || tplProp.PropInfo.VisAttribute.ShowInitially) { // Show also formulas with " " property.
                    property = new StudioProperty(template, tplProp, instance);
                    property.attr = tplProp.PropInfo.VisAttribute;

                    if (property.attr == null) { // Property without attribute will cause null-ref in Sort, etc. Give them a dummy attribute
                        Debug.Print(Tpl.Name + "! " + property.TemplateProperty.Name + " has no VisPropAtt");
                        property.attr = new VisPropAtt(VisPropAtt.PropertyType.Property, " ");
                        property.attr.Position = 0;
                    }
                    AddCategorizedItem(categories, "Event", property);
                }
            }

            // Look at all designer properties with or without a formula.
            var orderedDesignProps = from v in template.DesignProperties orderby v.Key ascending select v;
            foreach (var pair in orderedDesignProps) {
                property = new StudioProperty(template, pair.Value, instance);
                property.attr = new VisPropAtt(VisPropAtt.PropertyType.Property, " ");
                AddCategorizedItem(categories, "Designer", property);
            }

            List<CategoryData> compressed = new List<CategoryData>(10);
            for (int i = 0; i < categories.Count; i++)
                if (categories[i].Properties.Count > 0) compressed.Add(categories[i]);
            return compressed;
        }

        public static void AddCategorizedItem(List<CategoryData> categories, string categoryName, StudioProperty property) {
            CategoryData c = categories.Find(delegate(CategoryData i) { return i.Name == categoryName; });
            if (c == null) {
                c = new CategoryData(categoryName);
                categories.Add(c);
            }
            c.Properties.Add(property); // All template properties are added one by one. The same one will not be added twice.
        }

        private void UpdateCategories(List<CategoryData> categories) {
            //categories.Sort(delegate(CategoryData i1, CategoryData i2) { return i1.Name.CompareTo(i2.Name); });
            if (categories.Count < Categories.Count) {
                var index = categories.Count - 1;
                if (index < 0) index = 0;
                var count = Categories.Count - categories.Count;
                for (int i = 0; i < count; i++) {
                    var category = Categories[index + i];
                    category.expandButton.MouseDown -= Update;
                    category.hideButton.MouseDown -= Update;
                }
                Categories.RemoveRange(index, count);
            }

            for (int i = 0; i < categories.Count; i++) {
                Category category;
                if (i >= Categories.Count) {
                    category = new Category(categories[i].Name);
                    category.expandButton.MouseDown += Update;
                    category.hideButton.MouseDown += Update;
                    Categories.Add(category);
                } else {
                    category = Categories[i];
                    category.SetName(categories[i].Name);
                }
                UpdateProperties(category, categories[i]);
            }
        }

        private void UpdateProperties(Category category, CategoryData data) {
            if (data.Properties.Count < category.Properties.Count) {
                var index = data.Properties.Count - 1;
                if (index < 0) index = 0;
                var count = category.Properties.Count - data.Properties.Count;
                category.Properties.RemoveRange(index, count);
            }
            for (int i = 0; i < data.Properties.Count; i++) {
                if (i >= category.Properties.Count) {
                    var property = new PropertyGridLine(this, data.Properties[i]);
                    category.Properties.Add(property);
                } else {
                    var property = category.Properties[i];
                    property.UpdateData(data.Properties[i]);
                }
            }
            //data.Properties.Sort((a, b) => a.attr.Position.CompareTo(b.attr.Position)); Sorted elsewhere.
        }

        private void propertyBox_Loaded(object sender, RoutedEventArgs e) {

        }

        private void menu_Loaded(object sender, RoutedEventArgs e) {

        }
    }

    public class NameItem { // These are the items in autolistbox.
        public ImageSource Icon { get; set; }
        public string dotOperator { get; set; }
        public string Prefix { get; set; }
        public bool PrefixNeeded; // True if Prefix has to be inserted. (PrefixColor will be black in this case.)
        public string PrefixColor { get; set; }
        public string Name { get; set; } // Includes a dot and/or a table prefix when needed
        public bool insertPossible; // False for e.g. No Suggestion , ) + -
        public string Description { get; set; } // Will often be LongExplanation from the Attributes (annotations), but can be special descriptions.
        public string ToolTipText { get; set; }
        public int pos; // To allow sorting

        public NameItem(ImageSource icon, string dot, string prefix, bool prefixNeeded, string name, bool insertPossible, string description, string tooltipText, int pos) {
            this.Icon = icon;
            this.dotOperator = dot;
            this.Prefix = prefix;
            this.PrefixNeeded = prefixNeeded;
            this.PrefixColor = prefixNeeded? "Black" : "Gray";
            this.Name = name;
            this.insertPossible = insertPossible;
            this.Description = description;
            this.ToolTipText = tooltipText;
            this.pos = pos;
        }
    }

    public class CategoryData {
        public string Name;
        //public int CategoryID = 4; // For showing the categories in a non-alphabeticl sequence.
        public List<StudioProperty> Properties = new List<StudioProperty>();
        public CategoryData(string name) {
            this.Name = name;
        }
    }

    public class PropertyBoxManager {
        public static DockingManager dockManager;
        public static List<StudioGrid> StudioGrids;
        public static StudioGrid SelectedStudioGrid;
        public int standardDockWidth = 500;
        public int standardDockHeight = 500;

        public PropertyBoxManager(DockingManager dockingManager, List<StudioGrid> studioGrids) {
            dockManager = dockingManager;
            StudioGrids = studioGrids;

            TemplateHandler.TemplateSelected += TemplateSelected;
            TemplateHandler.InstanceSelected += InstanceSelected;
            //PropertyBoxMenu.SelectionChanged += MenuSelectionChanged;
            VisEngine.ShowForm += OnShowForm;

            foreach (var studioGrid in studioGrids) {
                studioGrid.Dock.IsActiveContentChanged += OnActiveContentChanged;
                studioGrid.Dock.StateChanged += OnGridStateChanged;
                studioGrid.Grid.LastUsed = DateTime.Now;
            }
        }

        private void OnGridStateChanged(object sender, RoutedEventArgs e) {
            var selectedDock = (DockableContent)sender;
            StudioGrid grid = null;
            for (int i = 0; i < StudioGrids.Count; i++) {
                if (StudioGrids[i].Dock != selectedDock) continue;
                grid = StudioGrids[i];
                break;
            }
            if (selectedDock.State == DockableContentState.DockableWindow) {
                if (selectedDock.Parent != null && selectedDock.Parent is FloatingDockablePane && (selectedDock.Parent as FloatingDockablePane).Parent != null && (selectedDock.Parent as FloatingDockablePane).Parent is DockableFloatingWindow) {
                    FloatingDockablePane parent = (FloatingDockablePane)selectedDock.Parent;
                    //parent.Width = 800;
                    DockableFloatingWindow grand = (DockableFloatingWindow)parent.Parent;
                    grand.Width = standardDockWidth;
                    grand.Height = standardDockHeight;
                }
            }
        }
        private void OnActiveContentChanged(object sender, EventArgs e) {
            var selectedDock = (DockableContent)sender;
            for(int i=0; i<StudioGrids.Count; i++){
                if (StudioGrids[i].Dock != selectedDock) continue;
                SelectedStudioGrid = StudioGrids[i];
                SelectedStudioGrid.Grid.LastUsed = DateTime.Now;
                return;
            }
        }

        //private void MenuSelectionChanged(object sender, EventArgs e) {
        //    foreach (var studioGrid in StudioGrids) {
        //        if (studioGrid.Grid.Tpl != null) {
        //            string clone = "";
        //            if (studioGrid.Grid.Tpl is FormTemplate) clone = " cloneID " + (studioGrid.Grid.Tpl as FormTemplate).cloneID.ToString(); 
        //            studioGrid.Dock.Title = studioGrid.Grid.Tpl.Name + clone;
        //        }

        //        //Debug.Print("Update menu: " + StudioBase.sw1.ElapsedMilliseconds);
        //        studioGrid.Grid.UpdateMenu();
        //        //Debug.Print("Update properties: " + StudioBase.sw1.ElapsedMilliseconds);
        //        studioGrid.Grid.UpdateProperties();
        //    }
        //}

        public void CloseGrids() {
            for (int i = 0; i < StudioGrids.Count; i++)
                StudioGrids[i].Dock.Close(true);
        }

        private void OnShowForm(object sender, DesignModeEventArgs e) {
            // Triggered by the run-time VisEngine at ShowForm, i.e. after Refresh and OpenForm. 
            // The run-time VisEngine changes only InstanceProp values and error messages, so nothing else needs update in Studio.
            var formInstance = (VisForm)e.form.Tag;
            formInstance.FormClosed -= OnFormsClosed;
            formInstance.FormClosed += OnFormsClosed;
            StudioBase.Studio.simulatedDateTextbox.Text = SystemItems.Now().ToString();
            StudioBase.Studio.Zoom.Text = SystemItems.Zoom2.ToString();

            foreach (StudioGrid studioGrid in StudioGrids) {
                UpdateInstanceValues(studioGrid.Grid);
            }

            // Old version where everything was updated
            foreach (var studioGrid in StudioGrids) {
                //if (studioGrid.Grid.Tpl != null)
                //    studioGrid.Dock.Title = studioGrid.Grid.Tpl.Name;
                studioGrid.Grid.UpdateMenu(); 
                //studioGrid.Grid.UpdateProperties(); // Unnecessary after OpenForm.
            }
        }

        private static void UpdateInstanceValues(PropertyBoxTemplate tpl) {
            if (tpl.Tpl == null) return; // Null grids are unused.
            foreach (Category c in tpl.Categories)
                foreach (PropertyGridLine prop in c.Properties) { // Lot of redundancy between property and property box.
                    prop.Property.visComponent = tpl.cmp;
                    prop.Property.controlAddress = new VisEngine.ControlAddress(prop.Property.visComponent);
                    prop.instanceValue.Text = prop.Property.GetVisComponentValue();
                    prop.instanceValue.ToolTip = prop.instanceValue.Text;
                    prop.instanceValue.TextWrapping = TextWrapping.NoWrap;
                    //prop.HighlightErrors(); // Done also in UpdateRichTexts. If done here, obstructs the optimization in UpdateRichTexts.
                }
        }

        public static void UpdateRichText(TplProperty tplProp) { 
            // Used by Execute and Undo to update the rich text formula and the error indications.
            foreach (var studioGrid in PropertyBoxManager.StudioGrids) {
                foreach (Category c in studioGrid.Grid.Categories)
                    foreach (PropertyGridLine prop in c.Properties) {
                        if (prop.Property.attr.ReadOnly) {
                            prop.SetRichTextFormula(prop.Property.attr.DefaultFormula, prop.Property.TemplateProperty, -1);
                            prop.formulaEditor.IsReadOnly = true;
                        } else {
                            if (prop.tagTokens.Count > 0 || prop.Property.TemplateProperty == tplProp)
                                prop.SetRichTextFormula(prop.Property.TemplateProperty.Formula, prop.Property.TemplateProperty, -1);
                            // Setting the RichTextFormula clears error indications.
                            prop.HighlightErrors(); // Sets the new error indications.
                        }
                    }
            }
        }

        public static void UpdateAllGrids(bool restructure) { /* Used by Execute and Undo to make a full update of the grids.
            When restructure == true, deletes all property lines and creates them from scratch. Needed when TemplateType or PropertyName is changed. 
            The kernel creates new properties in this case and the Grid will refer to deleted properties unless cleared.
            When undoing while edting a line, restructure somehow deletes the changes in the line, next inserts them again. */
            PropertyBoxTemplate lastGrid = null;
            foreach (var studioGrid in PropertyBoxManager.StudioGrids) {
                Template tpl = studioGrid.Grid.Tpl;
                VisComponent cmp = studioGrid.Grid.cmp;
                if (tpl != null) {
                    lastGrid = studioGrid.Grid;
                    if (restructure) studioGrid.Grid.Categories.Clear();
                    studioGrid.Grid.UpdateProperties();
                    //string clone = "";
                    //if (studioGrid.Grid.Tpl is FormTemplate) clone = " cloneID " + (studioGrid.Grid.Tpl as FormTemplate).cloneID.ToString();
                    //studioGrid.Dock.Title = studioGrid.Grid.Tpl.Name + clone;
                    studioGrid.Dock.Title = tpl.Name + (tpl.FormTpl.cloneID == 0 ? "" : "(" + tpl.FormTpl.cloneID.ToString() + ")")
                    + (cmp == null || cmp.Row == null ? "" : "[" + cmp.Index().ToString() + "]");
                }
            }
            if (lastGrid != null) {
                lastGrid.UpdateMenu();
            }
        }

        public static void UpdateAllValues() { // Used by FormInteraction to update Form position and size.
            foreach (var studioGrid in PropertyBoxManager.StudioGrids) {
                UpdateInstanceValues(studioGrid.Grid);
            }
        }

        public static StudioGrid FindStudioGrid(Template tpl) {
            foreach (var studioGrid in StudioGrids) {
                if (studioGrid.Dock.State != DockableContentState.Hidden) {
                    var grid = studioGrid.Grid;
                    var template = grid.Tpl;
                    if (template != null && tpl.Name == template.Name && tpl.FormTpl.Name == template.FormTpl.Name) { 
                        // Show only one clone version.
                        template.FormTpl = tpl.FormTpl; // Ensure that cloneId is carried on.
                        return studioGrid;
                    }
                }
            }
            return null;
        }

        // Quick Sort Algorithm
        public static void SortGridsByViewTime() {
            if (StudioGrids.Count > 1)
                SortGrids(0, StudioGrids.Count - 1);
        }

        private static void SortGrids(int left, int right) {
            int i = left, j = right;

            StudioGrid pivot = StudioGrids[(left + right) / 2];

            while (i <= j)
            {
                while (StudioGrids[i].Grid.LastUsed.CompareTo(pivot.Grid.LastUsed) < 0)
                    i++;
                while (StudioGrids[j].Grid.LastUsed.CompareTo(pivot.Grid.LastUsed) > 0)
                    j--;

                if (i <= j)
                {
                    StudioGrid tmp = StudioGrids[i];
                    StudioGrids[i] = StudioGrids[j];
                    StudioGrids[j] = tmp;
                    i++;
                    j--;
                }
            }

            // Recursive calls
            if (left < j)
                SortGrids(left, j);

            if (i < right)
                SortGrids(i, right);
        }

        public static StudioGrid GetStudioGrid() {

            //sort the grids by view time
            SortGridsByViewTime();
            // Find the grid according to this priority: 1- the grid that is shown without a template
            //                                           2- the grid that is hidden without a template
            //                                           3- the grid that is hidden with a template
            //                                           3- the grid that is shown with a template

            StudioGrid shownWithoutTpl = null, hiddenWithoutTpl = null, hiddenWithTpl = null, shownWithTpl = null;
            for(int i=0; i<StudioGrids.Count; i++) {
                var dock = StudioGrids[i].Dock;
                var grid = StudioGrids[i].Grid;
                if (shownWithoutTpl==null && dock.State != DockableContentState.Hidden && dock.State!=DockableContentState.DockableWindow && grid.Tpl == null) // shown without a template
                    shownWithoutTpl=StudioGrids[i];
                if (hiddenWithoutTpl == null && dock.State == DockableContentState.Hidden && grid.Tpl == null) // hidden without a template
                    hiddenWithoutTpl = StudioGrids[i];
                if (hiddenWithTpl == null && dock.State == DockableContentState.Hidden && grid.Tpl != null) // hidden with a template
                    hiddenWithTpl = StudioGrids[i];
                if (shownWithTpl == null && dock.State != DockableContentState.Hidden && dock.State != DockableContentState.DockableWindow && grid.Tpl != null)
                    shownWithTpl = StudioGrids[i];
            }

            return shownWithoutTpl != null ? shownWithoutTpl : hiddenWithoutTpl != null ? hiddenWithoutTpl : hiddenWithTpl != null ? hiddenWithTpl : shownWithTpl != null ? shownWithTpl : StudioGrids[0];
        }

        private void OnFormsClosed(object sender, EventArgs e) { // The clone has been closed, i.e. all Forms in its bundle have been closed.
            VisForm visFrm = sender as VisForm;
            FormTemplate frmTpl = visFrm.tpl as FormTemplate;
            List<FormTemplate> clones = frmTpl.clones;
            int openClones = 0;
            for (int i = 0; i < clones.Count; i++) if (clones[i] != null && clones[i].open) openClones++;

            int visibleGrids = 0;
            foreach (var studioGrid in StudioGrids)
                if (studioGrid.Dock.State != DockableContentState.Hidden) visibleGrids++;

            foreach (var studioGrid in StudioGrids)
                if (studioGrid.Dock.State != DockableContentState.Hidden) {
                    var grid = studioGrid.Grid;
                    if (grid.Tpl == null || frmTpl == grid.Tpl.FormTpl) {
                        if (openClones == 0) {
                            if (visibleGrids > 1) { // Don't hide the last visible grid or the layout will be gone.
                                //studioGrid.Dock.Close(true);
                                visibleGrids--;
                                studioGrid.Dock.Hide(); // = DockableContentState.Hidden;
                            } else {
                                studioGrid.Dock.Title = "";
                                studioGrid.Grid.Reset();
                            }
                        }
                    }
                }
        }

        //private bool StudioGridsClosed(StudioGrid grid) {
        //    foreach (var studioGrid in StudioGrids) {
        //        if (studioGrid == grid) continue;
        //        if (studioGrid.Dock.State != DockableContentState.Hidden)
        //            return false;
        //    }
        //    return true;
        //}

        private void InstanceSelected(object sender, InstanceEventArgs e) { // Called  by Hook, DesignClick.
            StudioBase.Studio.Focus();
            Template tpl = e.Instance == null ? (Template)sender : e.Instance.tpl;
            StudioGrid grid = e.grid;
            bool newGrid = grid == null || grid != null && grid.Dock.State != DockableContentState.DockableWindow;
            bool changeGrid = grid != null && grid.Dock.State == DockableContentState.DockableWindow;
            string addr = (tpl.FormTpl.cloneID == 0 ? "" : "(" + tpl.FormTpl.cloneID.ToString() + ")") 
                + (e.Instance == null || e.Instance.Row==null ? "" : "[" + e.Instance.Index().ToString() + "]");

            StudioGrid dg = FindStudioGrid(tpl);
            if (dg != null) {
                dg.Grid.cmp = e.Instance; // Change the selected instance.
                dg.Grid.Tpl = tpl; // Ensure that the cloneID is carried on.
                //dg.Grid.controladdress
                UpdateInstanceValues(dg.Grid);
                dg.Dock.Title = tpl.Name + addr;
                dg.Grid.UpdateMenu();
                //dg.Grid.CreateProperties(e.Instance);
                if (dg.Dock.State != DockableContentState.DockableWindow) // don't dock it back if it is a floating window
                    dg.Dock.Show();
                else
                    ((dg.Dock.Parent as FloatingDockablePane).Parent as DockableFloatingWindow).Focus();  //just focus it if it is a floating window
                return;
            }
            if (newGrid) {
                dg = GetStudioGrid();
                dg.Dock.Title = tpl.Name + addr;
                dg.Grid.UpdateMenu();
                if (e.Instance == null) dg.Grid.CreateProperties(tpl); else dg.Grid.CreateProperties(e.Instance);
                //dg.Dock.Show(dockManager, AnchorStyle.Top);
                Pane pane = dg.Dock.ContainerPane;
                if (pane!=null && pane.Items.Count >1){
                    pane.Items.Remove(dg.Dock);
                    pane.Items.Add(dg.Dock);
                }
                dg.Dock.Show(dockManager, AnchorStyle.Top);
            }
            if (changeGrid) {
                grid.Grid.Tpl = tpl;
                string clone = "";
                if (tpl is FormTemplate) clone = " cloneID " + (tpl as FormTemplate).cloneID.ToString();
                grid.Dock.Title = tpl.Name + addr;
                grid.Grid.UpdateMenu();
                grid.Grid.CreateProperties(e.Instance);
                grid.Dock.Show();
            }
        }

        public static void ShowError(Error error) {
            var tpl = error.tpl;
            if (tpl == null) return; // Happens at errors in a vism file. Cannot be shown as a formula.
            string clone = tpl.FormTpl.cloneID == 0 ? "" : "(" + tpl.FormTpl.cloneID.ToString() + ")";
            StudioGrid dg = FindStudioGrid(tpl);
            if (dg == null)
                dg = GetStudioGrid();
            dg.Dock.Title = tpl.Name + clone;
            dg.Grid.UpdateMenu();

            VisComponent cmp = null;
            if (tpl.RowBundle != null && tpl.RowBundle.Instances.Count > 0) {
                var bundle = tpl.RowBundle.Instances;
                    if (error.Index >= 0 && error.Index < tpl.RowBundle.Instances.Count) {
                        cmp = bundle[error.Index].cmp;
                        dg.Dock.Title += "[" + error.Index + "]";
                }
            }
            if (cmp != null)
                dg.Grid.CreateProperties(cmp);
            else
                dg.Grid.CreateProperties(tpl);
            dg.Grid.ShowError(error);
            // Reuse Grid: dg.Dock.Show();
            // New Grid:
            dg.Dock.Show(dockManager, AnchorStyle.Top);
        }

        private void TemplateSelected(object sender, TemplateEventArgs e) { // Called when a Form is opened.
            //StudioBase.Studio.Focus(); // Don't put Studio on top of the newly opened form.
            // Don't select the Form as a property grid. The designer may like to see some other component.

            var tpl = e.template;
            string clone = tpl.FormTpl.cloneID == 0 ? "" : "(" + tpl.FormTpl.cloneID.ToString() + ")";

            StudioGrid dg = FindStudioGrid(tpl);
            if (dg != null) {
                //if (tpl is FormTemplate) clone = " cloneID " + (tpl as FormTemplate).cloneID.ToString();
                dg.Dock.Title = tpl.Name + clone;
                dg.Grid.UpdateMenu();
                dg.Grid.CreateProperties(tpl);
                dg.Dock.Show();
            } else { // Use a new StudioGrid.
                dg = GetStudioGrid();
                //if (tpl is FormTemplate) clone = " cloneID " + (tpl as FormTemplate).cloneID.ToString();
                dg.Dock.Title = tpl.Name + clone;
                dg.Grid.UpdateMenu();
                dg.Grid.CreateProperties(tpl);
                dg.Dock.Show(dockManager, AnchorStyle.Top);
            }
        }
    }

    public class StudioGrid {
        public DockableContent Dock;
        public PropertyBoxTemplate Grid;

        public StudioGrid(DockableContent dock, PropertyBoxTemplate grid) {
            this.Dock = dock;
            this.Grid = grid;
        }
    }
}