﻿using System.Windows;
using System.Windows.Documents;

namespace uVisStudio.VisPropertyBox
{
    public partial class PropertyDialog : Window
    {
        public string PropertyName;
        public string Formula;
        public PropertyGridLine gridLine;

        public PropertyDialog() {
            InitializeComponent();
            this.grid.Background = ColorConstants.DialogWindowBackground();
        }

        public PropertyDialog(PropertyGridLine gridLine, double points) : this() {
            double fontSize = points * 96 / 72;
            this.InitializeComponent();
            this.gridLine = gridLine;

            StudioProperty property = gridLine.Property;
            this.PropertyName = property.TemplateProperty.Name;
            this.Formula = gridLine.GetRichText(); // property.GetFormula();
            var template = property.Template;
            var name = template.FormTpl.Name + ": " + template.Name;
            this.controlTextBox.Text = gridLine.instanceValue.Text;
            this.controlTextBox.FontSize = (double)fontSize;
            this.nameLabel.FontSize = (double)fontSize;
            Title = name;

            this.nameTextBox.Text = this.PropertyName;
            this.nameTextBox.FontSize = (double)fontSize;
            SetFormula(this.Formula);
            this.formulaTextBox.FontSize = (double)fontSize;
        }

        public void SetFormula(string formula) {
            var flowDoc = new FlowDocument();
            flowDoc.Blocks.Add(new Paragraph(new Run(formula)));
            this.formulaTextBox.Document = flowDoc;
        }

        private void OkClick(object sender, RoutedEventArgs e) { // From Property dialog box
            var formula = this.GetFormulaText();
            this.gridLine.SetNameEditor(this.nameTextBox.Text);
            this.gridLine.SetRichTextFormula(formula, this.gridLine.Property.TemplateProperty, -1);
            this.gridLine.Commit(false);

            this.Close();
        }

        private void CancelClick(object sender, RoutedEventArgs e) {
            this.Close();
        }

        private string GetFormulaText()
        {
            var text = new TextRange(this.formulaTextBox.Document.ContentStart, this.formulaTextBox.Document.ContentEnd).Text;
            text = text.TrimEnd();
            return text;
        }
    }
}
